from typing import List

import serial
from abc import ABC, abstractmethod

from .channel import Channel
import logging


class Module(ABC):
    def __init__(self, _serial: serial.Serial, _logger: logging.Logger, module: int):
        """Initialize the Module object.

        Args:
            _serial (serial.Serial): The serial object used for communication.
            module (int): The module number.

        """
        self._serial = _serial
        self._logger = _logger
        self._module = module
        self._channels: List[Channel] = []

    @property
    def module(self) -> int:
        """The module number.

        Returns:
            int: The module number.

        """
        return self._module

    @property
    @abstractmethod
    def number_of_channels(self) -> int:
        """The number of channels in the module.

        This property should be implemented by subclasses to provide the number of channels in the module.

        Returns:
            int: The number of channels.

        Raises:
            NotImplementedError: If the subclass does not implement this property.

        """
        pass

    @property
    @abstractmethod
    def channels(self) -> List[Channel]:
        """The channels in the module.

        This property should be implemented by subclasses to provide the channels in the module.

        Returns:
            List[Channel]: A list of Channel objects.

        Raises:
            NotImplementedError: If the subclass does not implement this property.

        """
        pass

    def channel(self, channel: int) -> Channel:
        """Get the specified channel in the module.

        Args:
            channel (int): The channel number.

        Returns:
            Channel: The Channel object.

        Raises:
            KeyError: If the channel number is invalid.

        """
        if channel not in range(len(self.channels)):
            raise KeyError(
                f"Invalid channel {channel}. Valid channels are 0..{self.number_of_channels - 1}"
            )
        return self.channels[channel]



from __future__ import annotations

from ..hvps import Hvps
from .module import Module
from ...commands.caen.channel import validate_board_number


class Caen(Hvps):
    def module(self, module: int = 0) -> Module:
        self._logger.debug(f"Getting module {module}")
        validate_board_number(module)
        if module not in self._modules:
            self._logger.debug(f"Creating module {module}")
            self._modules[module] = Module(self._serial, self._logger, module)
        return self._modules[module]



from __future__ import annotations

import serial
from serial.tools import list_ports
from typing import Dict
import logging
import uuid

from .module import Module


class Hvps:
    def __init__(
        self,
        baudrate: int = 115200,
        port: str | None = None,
        timeout: float | None = None,
        connect: bool = True,
        logging_level=logging.WARNING,
    ):
        """Initialize the HVPS (High-Voltage Power Supply) object.

        Args:
            baudrate (int, optional): The baud rate for serial communication. Defaults to 115200.
            port (str | None, optional): The serial port to use. If None, it will try to detect one automatically. Defaults to None.
            timeout (float | None, optional): The timeout for serial communication. Defaults to None.
            connect (bool, optional): Whether to connect to the serial port during initialization. Defaults to True.
            logging_level (int, optional): The logger level. Defaults to logger.WARNING.

        """

        self._logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}.{uuid.uuid4()}"
        )
        self._logger.setLevel(logging_level)

        formatter = logging.Formatter(
            "%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        self._logger.addHandler(stream_handler)

        self._modules: Dict[int, Module] = {}

        if port is None and connect:
            self._logger.info("No port specified, trying to detect one")
            ports = [port.device for port in list_ports.comports()]
            if len(ports) == 0:
                raise Exception("No ports available")
            port = ports[0]

        self._serial: serial.Serial = serial.Serial()
        self._serial.port = port
        self._logger.info(f"Using port {port}")
        self._serial.baudrate = baudrate
        self._logger.info(f"Using baud rate {self._serial.baudrate}")
        self._serial.timeout = timeout
        self._logger.debug(f"Using timeout {self._serial.timeout}")

        if connect:
            self._logger.debug("Opening serial port")
            self._serial.open()
            self._logger.debug("Serial port opened")

    def __del__(self):
        """Cleanup method to close the serial port when the HVPS object is deleted."""
        if hasattr(self, "_serial"):
            self._serial.close()

    def disconnect(self):
        """
        Disconnect from the serial port.
        """
        if not hasattr(self, "_serial"):
            return
        if self._serial.is_open:
            self._serial.close()

    @property
    def connected(self) -> bool:
        """
        Check if connected to the serial port.

        Returns:
            bool: True if connected, False otherwise.
        """
        return self._serial.is_open

    @property
    def port(self) -> str:
        """
        Get the serial port.

        Returns:
            str: The serial port.
        """
        return self._serial.port

    @property
    def baudrate(self) -> int:
        """
        Get the baud rate.

        Returns:
            int: The baud rate.
        """
        return self._serial.baudrate

    @property
    def timeout(self) -> float:
        """
        Get the timeout.

        Returns:
            float: The timeout.
        """
        return self._serial.timeout

    @property
    def serial(self):
        """
        Get the underlying serial.Serial object.

        Returns:
            serial.Serial: The serial.Serial object.
        """
        return self._serial

    @property
    def logger(self) -> logging.Logger:
        """
        Get the logger.

        Returns:
            logging.Logger: The logger.
        """
        return self._logger

    def set_logging_level(self, level: int):
        """
        Set the logging level.

        Args:
            level (int): The logging level.
        """
        self._logger.setLevel(level)

    def connect(self):
        """
        Connect to the serial port.
        """
        self._serial.open()

    # modules

    @property
    def modules(self):
        """
        Get the dictionary of modules.

        Returns:
            Dict[int, Module]: The dictionary of modules.
        """
        return self._modules

    def module(self, module: int = 0) -> Module:
        """Get the specified module.

        Args:
            module (int, optional): The module number. Defaults to 0.

        Returns:
            Module: The Module object.

        Raises:
            KeyError: If the module number is invalid.
        """
        if module not in self._modules:
            raise KeyError(f"Invalid module {module}")
        return self._modules[module]



"""
trip_action: 4
output_mode: 1
output_polarity: p
available_output_polarities: ['p', 'n']
voltage_set: 210.0
voltage_limit: 1955.292
voltage_nominal: 2000.0
available_output_modes: [1]
"""
import pytest
from typing import List, Set

from hvps.commands.iseg import _parse_response
from hvps.commands.iseg.channel import (
    _get_set_channel_command,
    _get_mon_channel_command,
)

from hvps.commands.iseg.module import (
    _get_set_module_command,
    _get_mon_module_command,
)


# TODO: validate commands generation and response parsing


def test_iseg_module_get_commands():
    with pytest.raises(ValueError):
        # invalid parameter name
        _get_mon_module_command("TEST")

    command = _get_mon_module_command(":READ:MODULE:EVENT:MASK")
    assert command == b":READ:MODULE:EVENT:MASK?\r\n"


def test_iseg_module_set_commands():
    with pytest.raises(ValueError):
        # invalid parameter name
        _get_set_module_command("TEST", 16)

    command = _get_set_module_command(":CONF:AVER", 16)
    assert command == b":CONF:AVER 16;*OPC?\r\n"


def test_iseg_channel_get_commands():
    with pytest.raises(ValueError):
        # invalid parameter name
        _get_mon_channel_command(0, "TEST")

    with pytest.raises(ValueError):
        # invalid channel number
        _get_mon_channel_command(-1, ":CONF:OUTPUT:POL:LIST")

    command = _get_mon_channel_command(0, ":CONF:OUTPUT:POL:LIST")
    assert command == b":CONF:OUTPUT:POL:LIST? (@0)\r\n"


def test_iseg_channel_set_commands():
    with pytest.raises(ValueError):
        # invalid parameter name
        _get_set_channel_command(0, "TEST", 0)

    with pytest.raises(ValueError):
        # invalid channel number
        _get_set_channel_command(-1, ":VOLT:BOUNDS", 10)

    command = _get_set_channel_command(0, ":VOLT:BOUNDS", 10)
    assert command == b":VOLT:BOUNDS 10,(@0);*OPC?\r\n"


def test_iseg_parse_response():
    response = b"1\r\n"
    parsed_response = _parse_response(response, None)
    assert parsed_response == ["1"]

    response = b"p,n\r\n"
    parsed_response = _parse_response(response, List[str])
    assert parsed_response == ["p", "n"]

    response = b"p\r\n"
    parsed_response = _parse_response(response, str)
    assert parsed_response == ["p"]

    response = b"1.23400E3\r\n"
    parsed_response = _parse_response(response, float)
    assert parsed_response == ["1.23400E3"]

    response = b"132\r\n"
    parsed_response = _parse_response(response, int)
    assert parsed_response == ["132"]

    # non-supported type
    with pytest.raises(ValueError):
        _parse_response(b"{1, 2}", Set[int])

    # type missmatch
    with pytest.raises(ValueError):
        _parse_response(b"1.23400E^3", float)



from __future__ import annotations
from typing import List

from ...commands.iseg.channel import (
    _get_set_channel_command,
    _get_mon_channel_command,
)
from ...commands.iseg import _write_command

from ..channel import Channel as BaseChannel


class Channel(BaseChannel):
    # Getters
    @property
    def trip_action(
        self,
    ) -> int:  # Instruction for NHR or SHR only. Instruction for NHS
        """
        Query the current action to be taken when a current trip occurs for the channel.

        Returns:
            The current action value.
        """
        command = _get_mon_channel_command(self._channel, ":CONF:TRIP:ACTION")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Unexpected response. Multiple action values received.")
        if not 0 <= int(response[0]) <= 4:
            raise ValueError(
                "Invalid action value. Expected values are 0, 1, 2, 3, or 4."
            )
        return int(response[0])

    @property
    def trip_timeout(
        self,
    ) -> int:  # Instruction for NHR or SHR only. Instruction for NHS
        """
        Query the current action to be taken when a current trip occurs for the channel.

        Returns:
            The current action value.
        """
        command = _get_mon_channel_command(self._channel, ":CONF:TRIP:TIME")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Unexpected response. Multiple action values received.")
        return int(response[0])

    @property
    def external_inhibit_action(
        self,
    ) -> int:  # Instruction for NHR or SHR only. Instruction for NHS
        """
        Query the current action to be taken when a current trip occurs for the channel.

        Returns:
            The current action value.
        """
        command = _get_mon_channel_command(self._channel, ":CONF:INHP:ACTION")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Unexpected response. Multiple action values received.")
        if not 0 <= int(response[0]) <= 4:
            raise ValueError(
                "Invalid action value. Expected values are 0, 1, 2, 3, or 4."
            )
        return int(response[0])

    @property
    def output_mode(self) -> int:  # Instruction for NHR or SHR only.
        """
        Query the configured channel output mode.

        Returns:
            The current output mode value.

        Output Mode allowed values: 1, 2, 3.
        """
        command = _get_mon_channel_command(self._channel, ":CONF:OUTPUT:MODE")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )

        if len(response) != 1:
            raise ValueError("Wrong number of values were sent, one value expected")

        allowed_modes = [1, 2, 3]
        if int(response[0]) not in allowed_modes:
            raise ValueError("Invalid output mode. Allowed modes are: 1, 2, 3.")
        return int(response[0])

    @property
    def available_output_modes(self) -> List[int]:  # Instruction for NHR or SHR only
        """
        Query the available channel output modes as a list.

        Returns:
            The list of available output mode values.

        """
        command = _get_mon_channel_command(self._channel, ":CONF:OUTPUT:MODE:LIST")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=List[int],
        )

        if 0 >= len(response) > 3:
            raise ValueError("Invalid number of output modes received.")

        try:
            output_values = [int(mode) for mode in response]
        except ValueError:
            raise ValueError("Invalid output mode value received.")

        allowed_modes = [1, 2, 3]
        for mode in output_values:
            if mode not in allowed_modes:
                raise ValueError(
                    f"Invalid output mode. Allowed modes are: {allowed_modes}"
                )

        return output_values

    @property
    def output_polarity(self) -> str:  # Instruction for NHR or SHR only
        """
        Query the current output polarity of the channel.

        Returns:
            The current output polarity. "p" for positive, "n" for negative.

        Example:
            polarity = channel.output_polarity
            print(polarity)  # Example output: "n"
        """
        command = _get_mon_channel_command(self._channel, ":CONF:OUTPUT:POL")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=str,
        )
        if len(response) != 1:
            raise ValueError("Last command hasn't been processed.")
        return response[0]

    @property
    def available_output_polarities(self):  # Instruction for NHR or SHR only
        """
        Query the available channel output polarities as a list.

        Returns:
            The list of available output polarity values.

        Output Polarity List:
            "p" - Positive
            "n" - Negative

        Example:
            polarities = channel.available_output_polarities
            print(polarities)  # Example output: ["p", "n"]
        """
        command = _get_mon_channel_command(self._channel, ":CONF:OUTPUT:POL:LIST")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=str,
        )

        if len(response) != 2:
            raise ValueError("Wrong number of values were sent, two values expected")

        for polarity in response:
            if polarity not in ["p", "n"]:
                raise ValueError(
                    "Invalid polarity value. Valid values are 'p' for positive and 'n' for negative."
                )

        return response

    @property
    def voltage_set(self):
        """
        Query the voltage set Vset in Volt.

        Returns:
            float: The voltage set Vset in Volt.

        Example:
            voltage = channel.voltage_set
            print(voltage)  # Example output: 1234.0
        """
        command = _get_mon_channel_command(self._channel, ":READ:VOLT")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=str,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were sent, one value expected")
        return float(response[0][:-1])

    @property
    def voltage_limit(self) -> float:  # Instruction for SHR only
        """
        Query the voltage limit Vlim in Volt.

        Returns:
            float: The voltage limit Vlim in Volt.

        Example:
            limit = channel.voltage_limit
            print(limit)  # Example output: 3000.0
        """
        command = _get_mon_channel_command(self._channel, ":READ:VOLT:LIM")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were sent, one value expected")
        return float(response[0][:-1])

    @property
    def voltage_nominal(self) -> float:
        """
        Query the channel voltage nominal Vnom in Volt.

        Returns:
            float: The channel voltage nominal Vnom in Volt.

        Example:
            nominal = channel.voltage_nominal
            print(nominal)  # Example output: 6000.0
        """
        command = _get_mon_channel_command(self._channel, ":READ:VOLT:NOM")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were sent, one value expected")
        return float(response[0][:-1])

    @property
    def voltage_mode(self) -> str:  # Instruction for NHR or SHR only
        """
        Query the configured channel voltage mode with polarity sign in Volt.

        Returns:
            str: The configured channel voltage mode with polarity sign in Volt.

        Example:
            mode = channel.voltage_mode
            print(mode)  # Example output: "6.0E3V"
        """
        command = _get_mon_channel_command(self._channel, ":READ:VOLT:MODE")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=str,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were sent, one value expected")
        return response[0][:-1]

    @property
    def voltage_mode_list(self) -> List[str]:  # Instruction for NHR or SHR only
        """
        Query the available channel voltage modes as a list.

        Returns:
            List[str]: The 3 available channel voltage modes as a list of strings.

        Example:
            mode_list = channel.voltage_mode_list
            print(mode_list)  # Example output: ["2.0E3V", "4.0E3V", "6.0E3V"]
        """
        command = _get_mon_channel_command(self._channel, ":READ:VOLT:MODE:LIST")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=List[str],
        )
        if len(response) != 3:
            raise ValueError("Wrong number of values were sent, three values expected")
        return [mode[:-1] for mode in response]

    @property
    def voltage_bounds(self) -> str:
        """
        Query the channel voltage bounds in Volt.

        Returns:
            str: The channel voltage bounds in Volt.

        Example:
            bounds = channel.voltage_bounds
            print(bounds)  # Example output: "0.00000E3V"
        """
        command = _get_mon_channel_command(self._channel, ":READ:VOLT:BOUNDS")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=str,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were sent, one value expected")
        return response[0][:-1]

    @property
    def set_on(self) -> bool:
        """
        Query the channel control bit "Set On".

        Returns:
            bool: True if the channel control bit is set on, False otherwise.

        Example:
            is_on = channel.set_on
            print(is_on)  # Example output: True
        """
        command = _get_mon_channel_command(self._channel, ":READ:VOLT:ON")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=bool,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0][:-1]) == 1

    @property
    def emergency_off(self) -> bool:
        """
        Query the channel control bit "Set Emergency Off".

        Returns:
            bool: True if the channel control bit is set to Emergency Off, False otherwise.

        Example:
            is_emergency_off = channel.emergency_off
            print(is_emergency_off)  # Example output: False
        """
        command = _get_mon_channel_command(self._channel, ":READ:VOLT:EMCY")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=bool,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0][:-1]) == 1

    @property
    def current_set(self) -> float:
        """
        Query the current set (Iset) in Ampere.

        Returns:
            float: The current set value.

        Example:
            current = channel.current_set
            print(current)  # Example output: 50.000E-6

        """
        command = _get_mon_channel_command(self._channel, ":READ:CURR")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-1])

    @property
    def current_limit(self) -> float:  # Instruction for SHR only
        """
        Query the current limit (Ilim) in Ampere.

        Returns:
            float: The current limit value.

        Example:
            limit = channel.current_limit
            print(limit)  # Example output: 5.00000E-3

        """
        command = _get_mon_channel_command(self._channel, ":READ:CURR:LIM")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(
            response[0][:-1]
        )  # Remove the last character from the response (unit)

    @property
    def current_nominal(self) -> float:
        """
        Query the current nominal (Inom) in Ampere.

        Returns:
            float: The current nominal value.

        Example:
            nominal = channel.current_nominal
            print(nominal)  # Example output: 6.00000E-3

        """
        command = _get_mon_channel_command(self._channel, ":READ:CURR:NOM")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(
            response[0][:-1]
        )  # Remove the last character from the response (unit)

    @property
    def current_mode(self) -> float:  # Instruction for NHR or SHR only
        """
        Query the configured channel current mode in Ampere.

        Returns:
            float: The current mode value.

        Example:
            mode = channel.current_mode
            print(mode)  # Example output: 2.00000E-3

        """
        command = _get_mon_channel_command(self._channel, ":READ:CURR:MODE")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(
            response[0][:-1]
        )  # Remove the last character from the response (unit)

    @property
    def current_mode_list(self) -> List[float]:  # Instruction for NHR or SHR only
        """
        Query the available channel current modes as a list.

        Returns:
            List[float]: The list of available current modes.

        Example:
            modes = channel.current_mode_list
            print(modes)  # Example output: [6.0E-3, 4.0E-3, 2.0E-3]

        """
        command = _get_mon_channel_command(self._channel, ":READ:CURR:MODE:LIST")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=List[float],
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        modes = [
            float(mode[:-1]) for mode in response
        ]  # Remove the last character from each mode (unit)
        return modes

    @property
    def current_bounds(self) -> float:
        """
        Query the channel current bounds in Ampere.

        Returns:
            float: The channel current bounds.

        Example:
            bounds = channel.current_bounds
            print(bounds)  # Example output: 0.00000E-3

        """
        command = _get_mon_channel_command(self._channel, ":READ:CURR:BOUNDS")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-1])

    @property
    def current_ramp_speed(self) -> float:  # Instruction for EHS, NHR or SHR only
        """
        Query the channel current ramp speed in Ampere/second.

        Returns:
            float: The channel current ramp speed.

        Example:
            speed = channel.current_ramp_speed
            print(speed)  # Example output: 2.0000E-3

        """
        command = _get_mon_channel_command(self._channel, ":READ:RAMP:CURR")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-3])

    @property
    def voltage_ramp_speed(self) -> float:  # Instruction for EHS, NHR or SHR only
        """
        Query the channel voltage ramp speed in Volt/second.

        Returns:
            float: The channel voltage ramp speed.

        Example:
            speed = channel.voltage_ramp_speed
            print(speed)  # Example output: 0.25000E3

        """
        command = _get_mon_channel_command(self._channel, ":READ:RAMP:VOLT")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-3])

    @property
    def voltage_ramp_speed_minimum(
        self,
    ) -> float:  # Instruction for EHS, NHR or SHR only
        """
        Query the channel voltage ramp speed minimum in Volt/second.

        Returns:
            float: The channel voltage ramp speed minimum.

        Example:
            speed_min = channel.voltage_ramp_speed_minimum
            print(speed_min)  # Example output: 0.00005E3

        """
        command = _get_mon_channel_command(self._channel, ":READ:RAMP:VOLT:MIN")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-3])

    @property
    def voltage_ramp_speed_maximum(
        self,
    ) -> float:  # Instruction for EHS, NHR or SHR only
        """
        Query the channel voltage ramp speed maximum in Volt/second.

        Returns:
            float: The channel voltage ramp speed maximum.

        Example:
            speed_max = channel.voltage_ramp_speed_maximum
            print(speed_max)  # Example output: 1.20000E3

        """
        command = _get_mon_channel_command(self._channel, ":READ:RAMP:VOLT:MAX")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-3])

    @property
    def current_ramp_speed_minimum(
        self,
    ) -> float:  # Instruction for EHS, NHR or SHR only
        """
        Query the channel current ramp speed minimum in Ampere/second.

        Returns:
            float: The channel current ramp speed minimum.

        Example:
            speed_min = channel.current_ramp_speed_minimum
            print(speed_min)  # Example output: 1.0000E-3

        """
        command = _get_mon_channel_command(self._channel, ":READ:RAMP:CURR:MIN")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-3])

    @property
    def current_ramp_speed_maximum(
        self,
    ) -> float:  # Instruction for EHS, NHR or SHR only
        """
        Query the channel current ramp speed maximum in Ampere/second.

        Returns:
            float: The channel current ramp speed maximum.

        Example:
            speed_max = channel.current_ramp_speed_maximum
            print(speed_max)  # Example output: 6.0000E-3

        """
        command = _get_mon_channel_command(self._channel, ":READ:RAMP:CURR:MAX")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-3])

    @property
    def channel_control(self) -> int:
        """
        Query the Channel Control register.

        Returns:
            int: The Channel Control register value.

        Example:
            control = channel.channel_control
            print(control)  # Example output: 8

        """
        command = _get_mon_channel_command(self._channel, ":READ:CHAN:CONTROL")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def channel_status(self) -> int:
        """
        Query the Channel Status register.

        Returns:
            int: The Channel Status register value.

        Example:
            status = channel.channel_status
            print(status)  # Example output: 132

        """
        command = _get_mon_channel_command(self._channel, ":READ:CHAN:STATUS")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def channel_event_mask(self) -> int:
        """
        Query the Channel Event Mask register.

        Returns:
            int: The Channel Event Mask register value.

        Example:
            mask = channel.channel_event_mask
            print(mask)  # Example output: 0
        """
        command = _get_mon_channel_command(self._channel, "READ:CHAN:EVENT:MASK")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def measured_voltage(self) -> float:
        """
        Query the measured channel voltage in Volt.

        Returns:
            float: The measured channel voltage.

        Example:
            voltage = channel.measured_voltage
            print(voltage)  # Example output: 1234.56
        """
        command = _get_mon_channel_command(self._channel, ":MEAS:VOLT")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-1])

    @property
    def measured_current(self) -> float:
        """
        Query the measured channel current in Ampere.

        Returns:
            float: The measured channel current.

        Example:
            current = channel.measured_current
            print(current)  # Example output: 0.00123456
        """
        command = _get_mon_channel_command(self._channel, ":MEAS:CURR")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-1])

    @property
    def channel_voltage_ramp_up_speed(
        self,
    ) -> float:  # Instruction for EHS, NHR or SHR only
        """
        Query the channel voltage ramp up speed in Volt/second.

        Returns:
            float: The channel voltage ramp up speed in Volt/second.

        Example:
            speed = channel.channel_voltage_ramp_up_speed
            print(speed)  # Example output: 0.250E3
        """
        command = _get_mon_channel_command(self._channel, ":CONF:RAMP:VOLT:UP")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:-3])  # Remove the last 3 characters (unit)

    @property
    def channel_voltage_ramp_down_speed(
        self,
    ) -> float:  # Instruction for EHS, NHR or SHR only
        """
        Query the channel voltage ramp down speed in Volt/second.

        Returns:
            float: The channel voltage ramp down speed in Volt/second.

        Example:
            speed = channel.channel_voltage_ramp_down_speed
            print(speed)  # Example output: 0.12500E3
        """
        command = _get_mon_channel_command(self._channel, ":CONF:RAMP:VOLT:DOWN")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(
            response[0][:-3]
        )  # Remove the last 3 characters (unit) and cast to float

    @property
    def channel_current_ramp_up_speed(self) -> float:
        """
        Query the channel current ramp up speed in Ampere/second.

        Returns:
            float: The channel current ramp up speed in Ampere/second.

        Example:
            speed = channel.channel_voltage_ramp_down_speed
            print(speed)  # Example output: 0.12500E3
        """
        command = _get_mon_channel_command(self._channel, ":CONF:RAMP:CURR:UP")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(
            response[0][:-3]
        )  # Remove the last 3 characters (unit) and cast to float

    @property
    def channel_current_ramp_down_speed(self) -> float:
        """
        Query the channel current ramp down speed in Ampere/second.

        Returns:
            float: The channel current ramp down speed in Ampere/second.

        """
        command = _get_mon_channel_command(self._channel, ":CONF:RAMP:CURR:DOWN")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(
            response[0][:-3]
        )  # Remove the last 3 characters (unit) and cast to float

    # Setters

    @trip_action.setter
    def trip_action(
        self, action: int
    ) -> None:  # Instruction for NHR or SHR only. Instruction for NHS
        """
        Set the action to be taken when a current trip occurs for the channel.

        Args:
            action: The action value to set. Possible values are:
                    - 0: No action (Trip status flag will be set after timeout)
                    - 1: Turn off the channel with ramp
                    - 2: Shut down the channel without ramp
                    - 3: Shut down the whole module without ramp
                    - 4: Disable the Delayed Trip function
        """
        if not 0 <= action <= 4:
            raise ValueError(
                "Invalid action value. Expected values are 0, 1, 2, 3, or 4."
            )
        command = _get_set_channel_command(self._channel, ":CONF:TRIP:ACTION", action)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    @trip_timeout.setter
    def trip_timeout(
        self, timeout: int
    ) -> None:  # Instruction for NHR or SHR only. Instruction for NHS
        """
        Set the trip timeout with one millisecond resolution.

        Args:
            timeout: The timeout value to set in milliseconds. Must be in the range 1 to 4095 ms.
        """
        if not 1 <= timeout <= 4095:
            raise ValueError("Timeout value must be in the range 1 to 4095 ms.")
        command = _get_set_channel_command(self._channel, ":CONF:TRIP:TIME", timeout)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    @external_inhibit_action.setter
    def external_inhibit_action(
        self, action: int
    ) -> None:  # Instruction for NHR or SHR only. Instruction for NHS
        """
        Set the action to be taken when an External Inhibit event occurs for the channel.

        Args:
            action: The action value to set. Possible values are:
                    - 0: No action (External Inhibit status flag will be set)
                    - 1: Turn off the channel with ramp
                    - 2: Shut down the channel without ramp
                    - 3: Shut down the whole module without ramp
                    - 4: Disable the External Inhibit function
        """
        if not 0 <= action <= 4:
            raise ValueError(
                "Invalid action value. Expected values are 0, 1, 2, 3, or 4."
            )
        command = _get_set_channel_command(self._channel, ":CONF:INHP:ACTION", action)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    @output_mode.setter
    def output_mode(self, mode: int) -> None:  # Instruction for NHR or SHR only.
        """
        Set the channel output mode.

        Args:
            mode: The output mode value to set.

        Raises:
            ValueError: If the specified mode value is not allowed.

        Output Mode allowed values: 1, 2, 3.
        """
        allowed_modes = [1, 2, 3]
        if mode not in allowed_modes:
            raise ValueError("Invalid output mode. Allowed modes are: 1, 2, 3.")
        command = _get_set_channel_command(self._channel, ":CONF:OUTPUT:MODE", mode)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    @output_polarity.setter
    def output_polarity(
        self, polarity: str
    ) -> None:  # Instruction for NHR or SHR only.
        """
        Set the output polarity of the channel.

        Args:
            polarity: The output polarity to set. Valid values are "p" for positive and "n" for negative.

        Raises:
            ValueError: If an invalid polarity value is provided.

        Example:
            channel.output_polarity("n")
        """
        if polarity not in ["p", "n"]:
            raise ValueError(
                "Invalid polarity value. Valid values are 'p' for positive and 'n' for negative."
            )
        command = _get_set_channel_command(self._channel, ":CONF:OUTPUT:POL", polarity)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @voltage_set.setter
    def voltage_set(self, vset: float) -> None:
        """
        Set the channel voltage set.

        Args:
            vset (float): The voltage set value to set in Volt.
        """
        command = _get_set_channel_command(self._channel, ":VOLT", vset)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @voltage_bounds.setter
    def voltage_bounds(self, vbounds: float) -> None:
        """
        Set the channel voltage bounds.

        Args:
            vbounds (float): The voltage bounds value to set in Volt.
        """
        command = _get_set_channel_command(self._channel, ":VOLT:BOUNDS", vbounds)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    def clear_emergency_off(self) -> None:
        """
        Clear the channel from state emergency off. The channel goes to state off.
        """
        command = _get_set_channel_command(self._channel, ":VOLT EMCY", "CLR")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @current_set.setter
    def current_set(self, iset: float) -> None:
        """
        Set the channel current set.

        Args:
            iset (float): The current set value to set in Ampere.
        """
        command = _get_set_channel_command(self._channel, ":CURR", iset)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    @current_bounds.setter
    def current_bounds(self, ibounds: float) -> None:
        """
        Set the channel current bounds.

        Args:
            ibounds (float): The current bounds value to set in Ampere.
        """
        command = _get_set_channel_command(self._channel, ":CURR:BOUNDS", ibounds)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    def set_channel_voltage_ramp_up_down_speed(
        self, speed: int
    ) -> None:  # Instruction for EHS, NHR or SHR only
        """
        Set the channel voltage ramp speed for up and down direction in Volt/second.

        Args:
            speed (int): The voltage ramp speed in Volt/second.

        Example:
            channel.set_channel_voltage_ramp_up_down_speed(250)
        """
        command = _get_set_channel_command(self._channel, ":CONF:RAMP:VOLT", speed)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @channel_voltage_ramp_up_speed.setter
    def channel_voltage_ramp_up_speed(
        self, speed: int
    ) -> None:  # Instruction for EHS, NHR or SHR only
        """
        Set the channel voltage ramp up speed in Volt/second.

        Args:
            speed (int): The voltage ramp up speed in Volt/second.

        Example:
            channel.set_channel_voltage_ramp_up_speed(250)
        """
        command = _get_set_channel_command(self._channel, ":CONF:RAMP:VOLT:UP", speed)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @channel_voltage_ramp_down_speed.setter
    def channel_voltage_ramp_down_speed(
        self, speed: float
    ) -> None:  # Instruction for EHS, NHR or SHR only
        """
        Set the channel voltage ramp down speed in Volt/second.

        Args:
            speed (float): The channel voltage ramp down speed to set in Volt/second.

        Example:
            channel.set_channel_voltage_ramp_down_speed(125.0)
        """
        command = _get_set_channel_command(self._channel, ":CONF:RAMP:VOLT:DOWN", speed)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    def set_channel_current_ramp_up_down_speed(
        self, speed: float
    ) -> None:  # Instruction for EHS, NHR or SHR only
        """
        Set the channel current ramp speed for up and down direction in Ampere/second.

        Args:
            speed (float): The channel current ramp down speed to set in Ampere/second.

        Example:
            channel.set_channel_current_ramp_up_down_speed(125.0)
        """
        command = _get_set_channel_command(self._channel, ":CONF:RAMP:CURR", speed)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @channel_current_ramp_up_speed.setter
    def channel_current_ramp_up_speed(
        self, speed: float
    ) -> None:  # Instruction for EHS, NHR or SHR only
        """
        Set the channel current ramp up speed in Ampere/second.

        Args:
            speed (float): The channel current ramp up speed to set in Ampere/second.

        Example:
            channel.set_channel_current_ramp_up_speed(125.0)
        """
        command = _get_set_channel_command(self._channel, ":CONF:RAMP:CURR:UP", speed)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @channel_current_ramp_down_speed.setter
    def channel_current_ramp_down_speed(
        self, speed: float
    ) -> None:  # Instruction for EHS, NHR or SHR only
        """
        Set the channel current ramp down speed in Ampere/second.

        Args:
            speed (float): The channel current ramp down speed to set in Ampere/second.

        Example:
            channel.set_channel_current_ramp_down_speed(125.0)
        """
        command = _get_set_channel_command(self._channel, ":CONF:RAMP:CURR:DOWN", speed)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    def switch_on_high_voltage(self) -> None:
        """
        Switch on the high voltage with the configured ramp speed.
        """
        command = _get_set_channel_command(self._channel, ":VOLT ON", None)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    def switch_off_high_voltage(self) -> None:
        """
        Switch off the high voltage with the configured ramp speed.
        """
        command = _get_set_channel_command(self._channel, ":VOLT OFF", None)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    def shutdown_channel_high_voltage(self) -> None:
        """
        Shut down the channel high voltage (without ramp). The channel stays in Emergency Off until the command EMCY CLR is given.
        """
        command = _get_set_channel_command(self._channel, ":VOLT EMCY OFF", None)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    def clear_channel_emergency_off(self) -> None:
        """
        Clear the channel from state emergency off. The channel goes to state off.
        """
        command = _get_set_channel_command(self._channel, ":VOLT EMCY CLR", None)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    def clear_event_status(self) -> None:
        """
        Clear the Channel Event Status register.
        """
        command = _get_set_channel_command(self._channel, ":EVENT CLEAR", None)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    def clear_event_bits(self, bits: int) -> None:
        """
        Clears single bits or bit combinations in the Channel Event Status register
        by writing a one to the corresponding bit position.

        Args:
            bits: The bits or bit combinations to clear. Should be provided as an integer.
        """
        command = _get_set_channel_command(self._channel, ":EVENT", bits)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")

    def set_event_mask(self, mask: int) -> None:
        """
        Set the Channel Event Mask register

        Args:
            mask: new mask value
        """
        command = _get_set_channel_command(self._channel, ":EVENT:MASK", mask)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if int(response[0]) != 1:
            raise ValueError("Last command haven't been processed.")



from __future__ import annotations
from functools import cached_property
from typing import List

from ...commands.iseg.module import _get_mon_module_command, _get_set_module_command
from ...commands.iseg import _write_command
from ...utils.utils import string_number_to_bit_array
from ..module import Module as BaseModule
from .channel import Channel


class Module(BaseModule):
    def channel(self, channel: int) -> Channel:
        return super().channel(channel)

    @cached_property
    def number_of_channels(self) -> int:
        """The number of channels in the module.

        Returns:
            int: The number of channels.
        """
        self._logger.debug("Getting number of channels")
        if not self._serial.is_open:
            # TODO: we should not cache the property if the serial port is not open
            self._logger.warning(
                "Serial port is not open. Returning 1 as number of channels."
            )
            return 1

        command = _get_mon_module_command(":READ:MODULE:CHANNELNUMBER")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def channels(self) -> List[Channel]:
        """The channels in the module.

        Returns:
            List[Channel]: A list of Channel objects.
        """
        if len(self._channels) == 0:
            self._logger.debug("Initializing channels")
            for channel in range(self.number_of_channels):
                self._logger.debug(f"Creating channel {channel}")
                self._channels.append(
                    Channel(ser=self._serial, channel=channel, logger=self._logger)
                )
        return self._channels

    @property
    def firmware_release(self) -> str:
        """
        Read out Firmware Release (XX.X)

        Returns:
            str: The firmware release.
        """
        command = _get_mon_module_command(":READ:FIRMWARE:RELEASE")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=str,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return response[0]

    @property
    def module_status(self) -> dict:
        """
        Read out module status register

        Returns:
            str: The board alarm status value.
        """
        command = _get_mon_module_command(":READ:MODULE:STATUS")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=dict,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")

        bit_array = string_number_to_bit_array(response[0])
        bit_array = list(reversed(bit_array))

        # TODO: review this
        return {
            "Is Voltage Ramp Speed Limited": bit_array[21],
            "Is Fast Ramp Down": bit_array[16],
            "Is Kill Enable": bit_array[15],
            "Is Temperature Good": bit_array[14],
            "Is Supply Good": bit_array[13],
            "Is Module Good": bit_array[12],
            "Is Event Active": bit_array[11],
            "Is Safety Loop Good": bit_array[10],
            "Is No Ramp": bit_array[9],
            "Is No Sum": bit_array[8],
            "Is Input Error": bit_array[6],
            "Is Service Is High": bit_array[4],
            "Voltage On": bit_array[3],
            "Is Fine Adjustment": bit_array[0],
        }

    @property
    def filter_averaging_steps(self) -> int:
        """Query the digital filter averaging steps.

        Returns:
            int: The number of steps for filtering.
        """
        command = _get_mon_module_command(":CONF:AVER")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def kill_enable(self) -> int:
        """Get the current value for the kill enable function.

        Returns:
            int: The current kill enable value. 1 for enable, 0 for disable.
        """
        command = _get_mon_module_command(":CONF:KILL")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def adjustment(self) -> int:
        """Get the fine adjustment state.

        Returns:
            int: The current fine adjustment state. 1 for on, 0 for off.
        """
        command = _get_mon_module_command(":CONF:ADJUST")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def module_can_address(self) -> int:
        """Query the module's CAN bus address.

        Returns:
            int: The current CAN bus address of the module.
        """
        command = _get_mon_module_command(":CONF:CAN:ADDR")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def module_can_bitrate(self) -> int:
        """Query the module's CAN bus bit rate.

        Returns:
            int: The current CAN bus bit rate of the module.
        """
        command = _get_mon_module_command(":CONF:CAN:BITRATE")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def serial_baud_rate(
        self,
    ) -> (
        int
    ):  # Instruction is currently implemented for EHS devices with serial interface only
        """Query the device's serial baud rate.

        Returns:
            int: The current serial baud rate of the device.
        """
        command = _get_mon_module_command(":CONF:SERIAL:BAUD")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def serial_echo_enable(self) -> int:
        """Check if serial echo is enabled or disabled.

        Returns:
            int: 1 if serial echo is enabled, 0 if disabled.
        """
        command = _get_mon_module_command(":CONF:SERIAL:ECHO")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def serial_echo_enabled(self) -> bool:
        return self.serial_echo_enable == 1

    @property
    def serial_echo_disabled(self) -> bool:
        return self.serial_echo_enable == 0

    @property
    def module_voltage_limit(self) -> float:
        """Query the module's voltage limit in percent.

        Returns:
            float: The current voltage limit of the module.
        """
        command = _get_mon_module_command(":READ:VOLT:LIM")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0].rstrip("%"))

    @property
    def module_current_limit(self) -> float:
        """Query the module's current limit in percent.

        Returns:
            float: The current current limit of the module.
        """
        command = _get_mon_module_command(":READ:CURR:LIM")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0].rstrip("%"))

    @property
    def module_voltage_ramp_speed(self) -> float:
        """Query the module's voltage ramp speed in percent/second.

        Returns:
            float: The current voltage ramp speed of the module.
        """
        command = _get_mon_module_command(":READ:RAMP:VOLT")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:3])

    @property
    def module_current_ramp_speed(self) -> float:
        """Query the module's current ramp speed in percent/second.

        Returns:
            float: The current current ramp speed of the module.
        """
        command = _get_mon_module_command(":READ:RAMP:CURR")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return float(response[0][:3])

    @property
    def module_control_register(self) -> int:
        """Query the Module Control register.

        Returns:
            int: The value of the Module Control register.
        """
        command = _get_mon_module_command(":READ:MODULE:CONTROL")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def module_status_register(self) -> int:
        """Query the Module Status register.

        Returns:
            int: The value of the Module Status register.
        """
        command = _get_mon_module_command(":READ:MODULE:STATUS")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def module_event_status_register(self) -> int:
        """Query the Module Event Status register.

        Returns:
            int: The value of the Module Event Status register.
        """
        command = _get_mon_module_command(":READ:MODULE:EVENT:STATUS")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def module_event_mask_register(self) -> int:
        """Query the Module Event Mask register.

        Returns:
            int: The value of the Module Event Mask register.
        """
        command = _get_mon_module_command(":READ:MODULE:EVENT:MASK")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def module_event_channel_status_register(self) -> int:
        """Query the Module Event Channel Status register.

        Returns:
            int: The value of the Module Event Channel Status register.
        """
        command = _get_mon_module_command(":READ:MODULE:EVENT:CHANSTAT")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def module_event_channel_mask_register(self) -> int:
        """Query the Module Event Channel Mask register.

        Returns:
            int: The value of the Module Event Channel Mask register.
        """
        command = _get_mon_module_command(":READ:MODULE:EVENT:CHANMASK")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def module_supply_voltage(self) -> List[float]:
        """Query the module supply voltages.

        Returns:
            Tuple[str, str, str, str, str, str]: The module supply voltages.
        """
        command = _get_mon_module_command(":READ:MODULE:SUPPLY? (@0-6)")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=List[float],
        )
        if len(response) != 7:
            raise ValueError("Wrong number of values were received, one value expected")

        return [float(string[:-1]) for string in response]

    @property
    def module_supply_voltage_p24v(self) -> float:
        """Query the module supply voltage +24 Volt.

        Returns:
            float: The module supply voltage +24 Volt.
        """
        command = _get_mon_module_command(":READ:MODULE:SUPPLY:P24V")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        voltage = response[0][:-1]
        return float(voltage)

    @property
    def module_supply_voltage_n24v(self) -> float:
        """Query the module supply voltage -24 Volt.

        Returns:
            float: The module supply voltage -24 Volt.
        """
        command = _get_mon_module_command(":READ:MODULE:SUPPLY:N24V")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        voltage = response[0][:-1]
        return float(voltage)

    @property
    def module_supply_voltage_p5v(self) -> float:
        """Query the module supply voltage +5 Volt.

        Returns:
            float: The module supply voltage +5 Volt.
        """
        command = _get_mon_module_command(":READ:MODULE:SUPPLY:P5V")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        voltage = response[0][:-1]
        return float(voltage)

    @property
    def module_supply_voltage_p3v(self) -> float:
        """Query the module internal supply voltage +3.3 Volt.

        Returns:
            float: The module internal supply voltage +3.3 Volt.
        """
        command = _get_mon_module_command(":READ:MODULE:SUPPLY:P3V")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        voltage = response[0][:-1]
        return float(voltage)

    @property
    def module_supply_voltage_p12v(self) -> float:
        """Query the module internal supply voltage +12 Volt.

        Returns:
            float: The module internal supply voltage +12 Volt.
        """
        command = _get_mon_module_command(":READ:MODULE:SUPPLY:P12V")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        voltage = response[0][:-1]
        return float(voltage)

    @property
    def module_supply_voltage_n12v(self) -> float:
        """Query the module internal supply voltage -12 Volt.

        Returns:
            float: The module internal supply voltage -12 Volt.
        """
        command = _get_mon_module_command(":READ:MODULE:SUPPLY:N12V")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        voltage = response[0][:-1]
        return float(voltage)

    @property
    def module_temperature(self) -> float:
        """Query the module temperature in degree Celsius.

        Returns:
            float: The module temperature in degree Celsius.
        """
        command = _get_mon_module_command(":READ:MODULE:TEMPERATURE")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=float,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        temperature = response[0][:-1]
        return float(temperature)

    @property
    def setvalue_changes_counter(self) -> int:
        """Query the setvalue changes counter.

        Returns:
            int: The setvalue changes counter.
        """
        command = _get_mon_module_command(":READ:MODULE:SETVALUE")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=int,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0])

    @property
    def firmware_name(self) -> str:
        """Query the module's firmware name.

        Returns:
            str: The firmware name.
        """
        command = _get_mon_module_command(":READ:FIRMWARE:NAME")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=str,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return response[0]

    @property
    def configuration_mode(self) -> bool:
        """Check if the device is in configuration mode.

        Returns:
            bool: true if in configuration mode, otherwise false.
        """
        command = _get_mon_module_command(":SYSTEM:USER:CONFIG")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=bool,
        )
        if len(response) != 1:
            raise ValueError("Wrong number of values were received, one value expected")
        return int(response[0]) == 1

    # Setters

    @serial_baud_rate.setter
    def serial_baud_rate(self, baud_rate: int) -> None:
        """Set the device's serial baud rate.

        Args:
            baud_rate (int): The serial baud rate to set.
        """
        command = _get_set_module_command(":CONF:SERIAL:BAUD", baud_rate)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != baud_rate:
            raise ValueError("Last command hasn't been processed.")

    @serial_echo_enable.setter
    def serial_echo_enable(self, enabled: int) -> None:
        """Enable or disable serial echo.

        Args:
            enabled (int): 1 to enable serial echo, 0 to disable.
        """
        if enabled not in [0, 1]:
            raise ValueError(
                "Invalid serial echo value. Please choose 1 for enabled or 0 for disabled."
            )

        command = _get_set_module_command(":CONF:SERIAL:ECHO", enabled)
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @filter_averaging_steps.setter
    def filter_averaging_steps(self, steps: int) -> None:
        """Set the number of digital filter averaging steps.

        Args:
            steps (int): The number of steps for filtering. Accepts values 1, 16, 64, 256, 512, or 1024.

        Raises:
            ValueError: If an invalid number of steps is provided.
        """
        valid_steps = [1, 16, 64, 256, 512, 1024]
        if steps not in valid_steps:
            raise ValueError(
                f"Invalid number of steps. Please choose from {valid_steps}."
            )

        command = _get_set_module_command(":CONF:AVER", str(steps))
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @kill_enable.setter
    def kill_enable(self, enable: int) -> None:
        """Set function kill enable (1) or kill disable (0).

        Args:
            enable (int): The kill enable value to set. Accepts 1 for enable or 0 for disable.
        """
        if enable not in [0, 1]:
            raise ValueError(
                "Invalid kill enable value. Please choose 1 for enable or 0 for disable."
            )

        command = _get_set_module_command(":CONF:KILL", str(enable))
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @adjustment.setter
    def adjustment(self, value: int) -> None:
        """Set the fine adjustment function on (1) or off (0).

        Args:
            value (int): The adjustment value to set. Accepts 1 for on or 0 for off.
        """
        if value not in [0, 1]:
            raise ValueError(
                "Invalid adjustment value. Please choose 1 for on or 0 for off."
            )

        command = _get_set_module_command(":CONF:ADJUST", str(value))
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @module_event_mask_register.setter
    def module_event_mask_register(self, mask: int) -> None:
        """Set the Module Event Mask register.

        Args:
            mask (int): The value to set in the Module Event Mask register.
        """
        command = _get_set_module_command(":CONF:EVENT:MASK", str(mask))
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @module_event_channel_mask_register.setter
    def module_event_channel_mask_register(self, mask: int) -> None:
        """Set the Module Event Channel Mask register.

        Args:
            mask (int): The value to set in the Module Event Channel Mask register.
        """
        command = _get_set_module_command(":CONF:EVENT:CHANMASK", str(mask))
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @module_can_address.setter
    def module_can_address(self, address: int) -> None:
        """Set the module's CAN bus address.

        Args:
            address (int): The CAN bus address to set (0-63).
        """
        if not (0 <= address <= 63):
            raise ValueError(
                "Invalid CAN bus address. Please choose an address between 0 and 63."
            )

        command = _get_set_module_command(":CONF:CAN:ADDR", str(address))
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    @module_can_bitrate.setter
    def module_can_bitrate(self, bitrate: int) -> None:
        """Set the module's CAN bus bit rate.

        Args:
            bitrate (int): The CAN bus bit rate to set. Accepts 125000 or 250000.
        """
        if bitrate not in [125000, 250000]:
            raise ValueError(
                "Invalid CAN bus bitrate. Please choose either 125000 or 250000."
            )

        command = _get_set_module_command(":CONF:CAN:BITRATE", str(bitrate))
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    def enter_configuration_mode(self, serial_number: int):
        """Set the device to configuration mode to change the CAN bitrate or address.

        Parameters:
            serial_number (int): The device serial number.

        """
        command = _get_set_module_command(":SYSTEM:USER:CONFIG", str(serial_number))
        _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )

    def exit_configuration_mode(self):
        """Set the device back to normal mode."""
        command = _get_set_module_command(":SYSTEM:USER:CONFIG", "0")
        _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )

    def set_serial_echo_enabled(self) -> None:
        """Enable serial echo."""
        self.serial_echo_enable = 1

    def set_serial_echo_disabled(self) -> None:
        """Disable serial echo."""
        self.serial_echo_enable = 0

    def reset_module_event_status(self) -> None:
        """Reset the Module Event Status register."""
        command = _get_set_module_command(":CONF:EVENT CLEAR", "")
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")

    def clear_module_event_status_bits(self, bits: int) -> None:
        """Clear single bits or bit combinations in the Module Event Status register.

        Args:
            bits (int): The bits to clear in the Module Event Status register.
        """
        command = _get_set_module_command(":CONF:EVENT", str(bits))
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            command=command,
            expected_response_type=None,
        )
        if len(response) != 1 or int(response[0]) != 1:
            raise ValueError("Last command hasn't been processed.")



from __future__ import annotations
from typing import List

import logging
import serial
import re


def _write_command(
    ser: serial.Serial,
    logger: logging.Logger,
    command: bytes,
    expected_response_type: type | None,
    response: bool = True,
) -> List[str] | None:
    logger.debug(f"Send command: {command}")
    ser.write(command)
    if not response:
        return None

    # echo reading
    response = ser.readline()
    logger.debug(f"Received response: {response}")
    if response != command:
        raise ValueError(
            f"Invalid handshake echo response: {response}. expected {command}"
        )

    # response reading
    response = ser.readline()
    response = _parse_response(response, expected_response_type)

    return response


def _parse_response(response: bytes, expected_response_type: type | None) -> List[str]:
    """Parse the response from a device.

    Args:
        response (bytes): The response received from the device.

    Returns:
        str: The parsed value extracted from the response.

    Raises:
        ValueError: If the response is invalid, cannot be decoded, or does not match the expected pattern.
    """

    try:
        response: str = response.decode("ascii").strip()
    except UnicodeDecodeError:
        raise ValueError(f"Invalid response: {response}")

    if expected_response_type == float or expected_response_type == List[float]:
        # pattern for a float in scientific notation followed or not by units
        pattern = (
            r"^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?(\s*[a-zA-Z]+(/+[a-zA-Z]+)?)?$"
        )
    elif expected_response_type == int or expected_response_type == List[int]:
        pattern = r"^[-+]?\d+$"
    elif expected_response_type == str or expected_response_type == List[str]:
        pattern = r"^[\x00-\x7F]+$"
    elif expected_response_type is None:
        return response.split(",")
    else:
        raise ValueError(
            f"expected value type of {response}, {expected_response_type}, is not float, int or str"
        )

    split_response = response.split(",")

    for r in split_response:
        match = re.match(pattern, r)
        if not match:
            raise ValueError(
                f"Invalid response: {response}, can't be identified as a {expected_response_type}, missmatch in RE"
            )

    return split_response



from hvps.devices.iseg.channel import Channel
from hvps.utils import get_serial_ports
from hvps import Iseg

import pytest
import sys
import serial
import logging

serial_port = "COM4"  # change this to the serial port you are using
serial_baud = 9600
timeout = 5.0


def serial_port_available():
    ports = get_serial_ports()
    return ports != []


def is_macos():
    return sys.platform == "Darwin"


serial_skip_decorator = pytest.mark.skipif(
    serial_port_available() or not is_macos(), reason="No serial ports available"
)


@serial_skip_decorator
def test_iseg_module_monitor():
    iseg = Iseg(
        port=serial_port,
        baudrate=serial_baud,
        connect=True,
        timeout=timeout,
        logging_level=logging.DEBUG,
    )

    print(
        f"Serial port status: connected: {iseg.connected}, port: {iseg.port}, baudrate: {iseg.baudrate}, timeout: {iseg.timeout}"
    )
    module = iseg.module(0)

    number_of_channels = module.number_of_channels
    print(f"number_of_channels: {number_of_channels}")

    firmware_release = module.firmware_release
    print(f"firmware_release: {firmware_release}")

    module_status = module.module_status
    print(f"module_status: {module_status}")

    filter_averaging_steps = module.filter_averaging_steps
    print(f"filter_averaging_steps: {filter_averaging_steps}")

    kill_enable = module.kill_enable
    print(f"kill_enable: {kill_enable}")

    adjustment = module.adjustment
    print(f"adjustment: {adjustment}")

    module_can_address = module.module_can_address
    print(f"module_can_address: {module_can_address}")

    module_can_bitrate = module.module_can_bitrate
    print(f"module_can_bitrate: {module_can_bitrate}")

    serial_echo_enabled = module.serial_echo_enabled
    print(f"serial_echo_enabled: {serial_echo_enabled}")

    serial_echo_disabled = module.serial_echo_disabled
    print(f"serial_echo_disabled: {serial_echo_disabled}")

    module_voltage_limit = module.module_voltage_limit
    print(f"module_voltage_limit: {module_voltage_limit}")

    module_current_limit = module.module_current_limit
    print(f"module_current_limit: {module_current_limit}")

    module_voltage_ramp_speed = module.module_voltage_ramp_speed
    print(f"module_voltage_ramp_speed: {module_voltage_ramp_speed}")

    module_current_ramp_speed = module.module_current_ramp_speed
    print(f"module_current_ramp_speed: {module_current_ramp_speed}")

    module_control_register = module.module_control_register
    print(f"module_control_register: {module_control_register}")

    module_status_register = module.module_status_register
    print(f"module_status_register: {module_status_register}")

    module_event_status_register = module.module_event_status_register
    print(f"module_event_status_register: {module_event_status_register}")

    module_event_mask_register = module.module_event_mask_register
    print(f"module_event_mask_register: {module_event_mask_register}")

    module_event_channel_status_register = module.module_event_channel_status_register
    print(
        f"module_event_channel_status_register: {module_event_channel_status_register}"
    )

    module_event_channel_mask_register = module.module_event_channel_mask_register
    print(f"module_event_channel_mask_register: {module_event_channel_mask_register}")

    module_supply_voltage = module.module_supply_voltage
    print(f"module_supply_voltage: {module_supply_voltage}")

    module_supply_voltage_p24v = module.module_supply_voltage_p24v
    print(f"module_supply_voltage_p24v: {module_supply_voltage_p24v}")

    module_supply_voltage_n24v = module.module_supply_voltage_n24v
    print(f"module_supply_voltage_n24v: {module_supply_voltage_n24v}")

    module_supply_voltage_p5v = module.module_supply_voltage_p5v
    print(f"module_supply_voltage_p5v: {module_supply_voltage_p5v}")

    module_supply_voltage_p3v = module.module_supply_voltage_p3v
    print(f"module_supply_voltage_p3v: {module_supply_voltage_p3v}")

    module_supply_voltage_p12v = module.module_supply_voltage_p12v
    print(f"module_supply_voltage_p12v: {module_supply_voltage_p12v}")

    module_supply_voltage_n12v = module.module_supply_voltage_n12v
    print(f"module_supply_voltage_n12v: {module_supply_voltage_n12v}")

    module_temperature = module.module_temperature
    print(f"module_temperature: {module_temperature}")

    setvalue_changes_counter = module.setvalue_changes_counter
    print(f"setvalue_changes_counter: {setvalue_changes_counter}")

    firmware_name = module.firmware_name
    print(f"firmware_name: {firmware_name}")

    configuration_mode = module.configuration_mode
    print(f"configuration_mode: {configuration_mode}")

    # Setter methods
    module.serial_baud_rate = 9600
    module.serial_echo_enable = 1
    with pytest.raises(ValueError):
        module.serial_echo_enable = 2

    module.filter_averaging_steps = 256
    with pytest.raises(ValueError):
        module.filter_averaging_steps = 257

    module.kill_enable = 1
    with pytest.raises(ValueError):
        module.kill_enable = 2

    module.adjustment = 0
    with pytest.raises(ValueError):
        module.adjustment = 2

    module.module_event_mask_register = 5

    module.module_can_address = 12
    with pytest.raises(ValueError):
        module.module_can_address = 250

    module.module_can_bitrate = 250000
    with pytest.raises(ValueError):
        module.module_can_bitrate = 250

    # Other methods
    module.enter_configuration_mode(12345)
    module.exit_configuration_mode()
    module.set_serial_echo_enabled()
    module.set_serial_echo_disabled()
    module.reset_module_event_status()
    module.clear_module_event_status_bits(8)


@serial_skip_decorator
def test_iseg_channel_monitor():
    ser = serial.Serial(serial_port, serial_baud, timeout=timeout)

    channel = Channel(ser, 0)

    trip_action = channel.trip_action
    print(f"trip_action: {trip_action}")

    output_mode = channel.output_mode
    print(f"output_mode: {output_mode}")

    output_polarity = channel.output_polarity
    print(f"output_polarity: {output_polarity}")

    available_output_polarities = channel.available_output_polarities
    print(f"available_output_polarities: {available_output_polarities}")

    voltage_set = channel.voltage_set
    print(f"voltage_set: {voltage_set}")

    voltage_limit = channel.voltage_limit
    print(f"voltage_limit: {voltage_limit}")

    voltage_nominal = channel.voltage_nominal
    print(f"voltage_nominal: {voltage_nominal}")

    available_output_modes = channel.available_output_modes
    print(f"available_output_modes: {available_output_modes}")

    voltage_mode = channel.voltage_mode
    print(f"voltage_mode: {voltage_mode}")

    voltage_mode_list = channel.voltage_mode_list
    print(f"voltage_mode_list: {voltage_mode_list}")

    voltage_bounds = channel.voltage_bounds
    print(f"voltage_bounds: {voltage_bounds}")

    set_on = channel.set_on
    print(f"set_on: {set_on}")

    emergency_off = channel.emergency_off
    print(f"emergency_off: {emergency_off}")

    current_set = channel.current_set
    print(f"current_set: {current_set}")

    current_limit = channel.current_limit
    print(f"current_limit: {current_limit}")

    current_nominal = channel.current_nominal
    print(f"current_nominal: {current_nominal}")

    current_mode = channel.current_mode
    print(f"current_mode: {current_mode}")

    modes = channel.current_mode_list
    print(f"modes: {modes}")

    bounds = channel.current_bounds
    print(f"bounds: {bounds}")

    speed = channel.current_ramp_speed
    print(f"speed: {speed}")

    speed = channel.voltage_ramp_speed
    print(f"speed: {speed}")

    speed_min = channel.voltage_ramp_speed_minimum
    print(f"speed_min: {speed_min}")

    speed_max = channel.voltage_ramp_speed_maximum
    print(f"speed_max: {speed_max}")

    speed_min = channel.current_ramp_speed_minimum
    print(f"speed_min: {speed_min}")

    speed_max = channel.current_ramp_speed_maximum
    print(f"speed_max: {speed_max}")

    control = channel.channel_control
    print(f"control: {control}")

    status = channel.channel_status
    print(f"status: {status}")

    mask = channel.channel_event_mask
    print(f"mask: {mask}")

    voltage = channel.measured_voltage
    print(f"voltage: {voltage}")

    current = channel.measured_current
    print(f"current: {current}")

    speed = channel.channel_voltage_ramp_up_speed
    print(f"speed: {speed}")

    speed = channel.channel_voltage_ramp_down_speed
    print(f"speed: {speed}")

    speed = channel.channel_current_ramp_up_speed
    print(f"speed: {speed}")

    speed = channel.channel_current_ramp_down_speed
    print(f"speed: {speed}")

    # Set the action to be taken when a current trip occurs for the channel
    channel.trip_action = 1
    with pytest.raises(ValueError):
        channel.trip_action = 5

    # Set the trip timeout in milliseconds
    channel.trip_timeout = 2000
    with pytest.raises(ValueError):
        channel.trip_timeout = 0

    # Set the action to be taken when an External Inhibit event occurs for the channel
    channel.external_inhibit_action = 2
    with pytest.raises(ValueError):
        channel.external_inhibit_action = 5

    # Set the channel output mode
    channel.output_mode = 3
    with pytest.raises(ValueError):
        channel.output_mode = 5

    # Set the output polarity of the channel
    channel.output_polarity = "n"
    with pytest.raises(ValueError):
        channel.output_polarity = "x"

    # Set the channel voltage set
    channel.voltage_set = 12.5

    # Set the channel voltage bounds
    channel.voltage_bounds = 10.0

    # Clear the channel from state emergency off
    channel.clear_emergency_off()

    # Set the channel current set
    channel.current_set = 2.5

    # Set the channel current bounds
    channel.current_bounds = 2.0

    # Set the channel voltage ramp speed for up and down direction
    channel.set_channel_voltage_ramp_up_down_speed(250)

    # Set the channel voltage ramp up speed
    channel.channel_voltage_ramp_up_speed = 200

    # Set the channel voltage ramp down speed
    channel.channel_voltage_ramp_down_speed = 150.0

    # Switch on the high voltage with the configured ramp speed
    channel.switch_on_high_voltage()

    # Switch off the high voltage with the configured ramp speed
    channel.switch_off_high_voltage()

    # Shut down the channel high voltage (without ramp)
    channel.shutdown_channel_high_voltage()

    # Clear the channel from state emergency off
    channel.clear_channel_emergency_off()

    # Clear the Channel Event Status register
    channel.clear_event_status()

    # Clears single bits or bit combinations in the Channel Event Status register
    channel.clear_event_bits(3)

    # Set the Channel Event Mask register
    channel.set_event_mask(7)



from __future__ import annotations

_mon_module_commands = {
    ":READ:MODULE:CHANNELNUMBER": "The number of channels in the module.",
    ":READ:FIRMWARE:RELEASE": "Read out Firmware Release (XX.X)",
    ":READ:MODULE:STATUS": "Read out module status register",
    ":CONF:AVER": "Query the digital filter averaging steps.",
    ":CONF:KILL": "Get the current value for the kill enable function.",
    ":CONF:ADJUST": "Get the fine adjustment state.",
    ":CONF:CAN:ADDR": "Query the module's CAN bus address.",
    ":CONF:CAN:BITRATE": "Query the module's CAN bus bit rate.",
    ":CONF:SERIAL:BAUD": "Query the device's serial baud rate.",
    ":CONF:SERIAL:ECHO": "Check if serial echo is enabled or disabled.",
    ":READ:VOLT:LIM": "Query the module's voltage limit in percent.",
    ":READ:CURR:LIM": "Query the module's current limit in percent.",
    ":READ:RAMP:VOLT": "Query the module's voltage ramp speed in percent/second.",
    ":READ:RAMP:CURR": "Query the module's current ramp speed in percent/second.",
    ":READ:MODULE:CONTROL": "Query the Module Control register",
    ":READ:MODULE:EVENT:STATUS": "Query the Module Event Status register",
    ":READ:MODULE:EVENT:MASK": "Query the Module Event Mask register",
    ":READ:MODULE:EVENT:CHANSTAT": "Query the Module Event Channel Status register",
    ":READ:MODULE:EVENT:CHANMASK": "Query the Module Event Channel Mask register",
    ":READ:MODULE:SUPPLY? (@0-6)": "Query the module supply voltages",
    ":READ:MODULE:SUPPLY:P24V": "Query the module supply voltage +24 Volt",
    ":READ:MODULE:SUPPLY:N24V": "Query the module supply voltage -24 Volt",
    ":READ:MODULE:SUPPLY:P5V": "Query the module supply voltage +5 Volt",
    ":READ:MODULE:SUPPLY:P3V": "Query the module internal supply voltage +3.3 Volt",
    ":READ:MODULE:SUPPLY:P12V": "Query the module internal supply voltage +12 Volt",
    ":READ:MODULE:SUPPLY:N12V": "Query the module internal supply voltage -12 Volt",
    ":READ:MODULE:TEMPERATURE": "Query the module temperature in degree Celsius",
    ":READ:MODULE:SETVALUE": "Query the setvalue changes counter",
    ":READ:FIRMWARE:NAME": "Query the module's firmware name",
    ":CONF:EVENT:MASK": "Query the Module Event Mask register",
    ":CONF:EVENT:CHANMASK": "Query the Module Event Channel Mask register",
}

_set_module_commands = {
    ":CONF:AVER": "Query the digital filter averaging steps.",
    ":CONF:KILL": "Get the current value for the kill enable function.",
    ":CONF:ADJUST": "Get the fine adjustment state.",
    ":CONF:CAN:ADDR": "Query the module's CAN bus address.",
    ":CONF:CAN:BITRATE": "Query the module's CAN bus bit rate.",
    ":CONF:SERIAL:BAUD": "Query the device's serial baud rate.",
    ":CONF:SERIAL:ECHO": "Check if serial echo is enabled or disabled.",
    ":CONF:EVENT:MASK": "Set the Module Event Mask register",
    ":CONF:EVENT:CHANMASK": "Set the Module Event Channel Mask register",
    ":SYSTEM:USER:CONFIG": "Set the device to configuration mode to change the CAN bitrate or address",
    ":CONF:EVENT CLEAR": "Reset the Module Event Status register",
    ":CONF:EVENT": "Clear single bits or bit combinations in the Module Event Status register",
}


def _get_mon_module_command(command: str) -> bytes:
    """
    Generates a query command string for monitoring a specific module.

    Args:
        command (str): The base command without the query symbol.

    Returns:
        bytes: The query command string as bytes.

    Raises:
        ValueError: If the provided command is not a valid command.

    Example:
        command = ":MEAS:CURR"
        query_command = _get_mon_module_command(command)
        print(query_command)
        b':MEAS:CURR?\r\n'
    """
    command = command.upper()
    if command not in _mon_module_commands:
        valid_commands = ", ".join(_mon_module_commands.keys())
        raise ValueError(
            f"Invalid command '{command}'. Valid commands are: {valid_commands}."
        )

    return f"{command.strip()}?\r\n".encode("ascii")


def _get_set_module_command(command: str, value: str | int | float | None) -> bytes:
    """
    Generates an order command as a bytes object to set a value for a specific module.

    Args:
        command (str): The base command without the value and channel suffix.
        value (str | int | float | None): The value to be set. Can be a string, integer, float, or None.

    Returns:
        bytes: The order command as a bytes object.

    Raises:
        ValueError: If the provided command is not a valid command.

    Example:
        command = ":VOLT"
        value = 200
        order_command = _get_set_module_command(command, value)
        print(order_command)
        b':VOLT 200;*OPC?\r\n'
    """
    command = command.upper()
    if command not in _set_module_commands:
        valid_commands = ", ".join(_set_module_commands.keys())
        raise ValueError(
            f"Invalid command '{command}'. Valid commands are: {valid_commands}."
        )
    if isinstance(value, float):
        value = f"{value:E}"

    return f"{command.strip()} {value};*OPC?\r\n".encode("ascii")



from hvps.utils import string_number_to_bit_array


def test_string_number_to_bit_array():
    for string, bit_array in [
        (
            "00003",
            [
                True,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        ),
        (
            "00000",
            [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        ),
        (
            "00001",
            [
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        ),
        (
            "01028",
            [
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        ),
    ]:
        assert string_number_to_bit_array(string) == bit_array



from __future__ import annotations

_mon_channel_commands = {
    ":CONF:TRIP:ACTION": "Query the current action to be taken when a current trip occurs for the channel.",
    ":CONF:TRIP:TIME": "Query the current action to be taken when a current trip occurs for the channel.",
    ":CONF:INHP:ACTION": "Query the current action to be taken when a current trip occurs for the channel.",
    ":CONF:OUTPUT:MODE": "Query the configured channel output mode.",
    ":CONF:OUTPUT:MODE:LIST": "Query the available channel output modes as a list.",
    ":CONF:OUTPUT:POL": "Query the current output polarity of the channel.",
    ":CONF:OUTPUT:POL:LIST": "Query the available channel output polarities as a list.",
    ":READ:VOLT": "Query the voltage set Vset in Volt.",
    ":READ:VOLT:LIM": "Query the voltage limit Vlim in Volt.",
    ":READ:VOLT:NOM": "Query the channel voltage nominal Vnom in Volt.",
    ":READ:VOLT:ON": "Query the channel control bit 'Set On'.",
    ":READ:VOLT:EMCY": "Query the channel control bit 'Set Emergency Off'.",
    ":READ:CURR": "Query the current set (Iset) in Ampere.",
    ":READ:CURR:LIM": "Query the current limit (Ilim) in Ampere.",
    ":READ:CURR:NOM": "Query the current nominal (Inom) in Ampere.",
    ":READ:CURR:MODE": "Query the configured channel current mode in Ampere.",
    ":READ:CURR:MODE:LIST": "Query the available channel current modes as a list.",
    ":READ:CURR:BOUNDS": "Query the channel current bounds in Ampere.",
    ":READ:RAMP:CURR": "Query the channel current ramp speed in Ampere/second.",
    ":READ:RAMP:VOLT": "Query the channel voltage ramp speed in Volt/second.",
    ":READ:RAMP:VOLT:MIN": "Query the channel voltage ramp speed minimum in Volt/second.",
    ":READ:RAMP:VOLT:MAX": "Query the channel voltage ramp speed maximum in Volt/second.",
    ":READ:RAMP:CURR:MIN": "Query the channel current ramp speed minimum in Ampere/second.",
    ":READ:RAMP:CURR:MAX": "Query the channel current ramp speed maximum in Ampere/second.",
    ":READ:CHAN:CONTROL": "Query the Channel Control register.",
    ":READ:CHAN:STATUS": "Query the Channel Status register.",
    ":READ:CHAN:EVENT:MASK": "Query the Channel Event Mask register.",
    ":MEAS:VOLT": "Query the measured channel voltage in Volt.",
    ":MEAS:CURR": "Query the measured channel current in Ampere.",
    ":CONF:RAMP:VOLT:UP": "Query the channel voltage ramp up speed in Volt/second.",
    ":CONF:RAMP:VOLT:DOWN": "Query the channel voltage ramp down speed in Volt/second.",
    ":CONF:RAMP:CURR:UP": "Query the channel current ramp up speed in Ampere/second.",
    ":CONF:RAMP:CURR:DOWN": "Query the channel current ramp down speed in Ampere/second.",
}

_set_channel_commands = {
    ":CONF:TRIP:ACTION": "Set the action to be taken when a current trip occurs for the channel.",
    ":CONF:TRIP:TIME": "Set the trip timeout with one millisecond resolution.",
    ":CONF:INHP:ACTION": "Set the action to be taken when an External Inhibit event occurs for the channel.",
    ":CONF:OUTPUT:MODE": "Set the channel output mode.",
    ":CONF:OUTPUT:POL": "Set the output polarity of the channel.",
    ":VOLT": "Set the channel voltage set.",
    ":VOLT:BOUNDS": "Set the channel voltage bounds.",
    ":VOLT EMCY": "Clear the channel from state emergency off.",
    ":CURR": "Set the channel current set.",
    ":CURR:BOUNDS": "Set the channel current bounds in Ampere.",
    ":CONF:RAMP:VOLT": "Set the channel voltage ramp speed for up and down direction in Volt/second.",
    ":CONF:RAMP:VOLT:UP": "Set the channel voltage ramp up speed in Volt/second.",
    ":CONF:RAMP:VOLT:DOWN": "Set the channel voltage ramp down speed in Volt/second.",
    ":CONF:RAMP:CURR": "Set the channel current ramp speed for up and down direction in Ampere/second.",
    ":CONF:RAMP:CURR:UP": "Set the channel current ramp up speed in Ampere/second.",
    ":CONF:RAMP:CURR:DOWN": "Set the channel current ramp down speed in Ampere/second.",
    ":VOLT ON": "Switch on the high voltage with the configured ramp speed.",
    ":VOLT OFF": "Switch off the high voltage with the configured ramp speed.",
    ":VOLT EMCY OFF": "Shut down the channel high voltage (without ramp) or clear the channel from state emergency off.",
    ":VOLT EMCY CLR": "Clear the channel from state emergency off. The channel goes to state off.",
    ":EVENT CLEAR": "Clear the Channel Event Status register.",
    ":EVENT": "Clears single bits or bit combinations in the Channel Event Status register by writing a one to the corresponding bit position.",
    ":EVENT:MASK": "Clears single bits or bit combinations in the Channel Event Status register by writing a one to the corresponding bit position.",
}


def _get_mon_channel_command(channel: int, command: str) -> bytes:
    """
    Generates a query command string for monitoring a specific channel.

    Args:
        channel (int): The channel number.
        command (str): The base command without the query symbol.

    Returns:
        bytes: The query command string as bytes.

    Raises:
        ValueError: If the provided channel is not a valid positive integer.
        ValueError: If the provided command is not a valid command.

    Example:
        command = ":MEAS:CURR"
        channel = 3
        query_command = _get_mon_channel_command(channel, command)
        print(query_command)
        b':MEAS:CURR? (@3)\r\n'
    """
    if channel < 0:
        raise ValueError(
            f"Invalid channel '{channel}'. Valid channels are positive integers."
        )
    command = command.upper()
    if command not in _mon_channel_commands:
        valid_commands = ", ".join(_mon_channel_commands.keys())
        raise ValueError(
            f"Invalid command '{command}'. Valid commands are: {valid_commands}."
        )

    return f"{command.strip()}? (@{channel})\r\n".encode("ascii")


def _get_set_channel_command(
    channel: int, command: str, value: str | int | float | None
) -> bytes:
    """
    Generates an order command as a bytes object to set a value for a specific channel.

    Args:
        channel (int): The channel number.
        command (str): The base command without the value and channel suffix.
        value (str | int | float | None): The value to be set. Can be a string, integer, float, or None.

    Returns:
        bytes: The order command as a bytes object.

    Raises:
        ValueError: If the provided channel is not a valid positive integer.
        ValueError: If the provided command is not a valid command.

    Example:
        channel = 4
        command = ":VOLT"
        value = 200
        order_command = _get_set_channel_command(channel, command, value)
        print(order_command)
        b':VOLT 200,(@4);*OPC?\r\n'
    """
    if channel < 0:
        raise ValueError(
            f"Invalid channel '{channel}'. Valid channels are positive integers."
        )
    command = command.upper()
    if command not in _set_channel_commands:
        valid_commands = ", ".join(_mon_channel_commands.keys())
        raise ValueError(
            f"Invalid command '{command}'. Valid commands are: {valid_commands}."
        )

    if isinstance(value, float):
        value = f"{value:E}"

    return f"{command.strip()} {value},(@{channel});*OPC?\r\n".encode("ascii")



import sys
import pytest

from hvps import Caen
from hvps.utils import get_serial_ports

import logging

# find a way to only run these tests if a serial port connection exists

serial_port = "COM4"  # change this to the serial port you are using
serial_baud = 115200
timeout = 5.0


def serial_port_available():
    ports = get_serial_ports()
    return ports != []


def is_macos():
    return sys.platform == "Darwin"


serial_skip_decorator = pytest.mark.skipif(
    serial_port_available() or not is_macos(), reason="No serial ports available"
)


@serial_skip_decorator
def test_caen_module_monitor():
    # no ports available
    caen = Caen(
        port=serial_port,
        baudrate=serial_baud,
        connect=True,
        timeout=timeout,
        logging_level=logging.DEBUG,
    )
    print(
        f"Serial port status: connected: {caen.connected}, port: {caen.port}, baudrate: {caen.baudrate}, timeout: {caen.timeout}"
    )
    module = caen.module(0)

    control_mode = module.control_mode
    print(f"Control mode: {control_mode}")

    interlock_mode = module.interlock_mode
    print(f"Interlock mode: {interlock_mode}")

    name = module.name
    print(f"Name: {name}")

    bd = module.bd
    print(f"BD value: {bd}")

    number_of_channels = module.number_of_channels
    print(f"Number of channels: {number_of_channels}")

    firmware_release = module.firmware_release
    print(f"Firmware release: {firmware_release}")

    serial_number = module.serial_number
    print(f"Serial number: {serial_number}")

    interlock_status = module.interlock_status
    print(f"Interlock status: {interlock_status}")

    interlock_open = module.interlock_open
    print(f"Interlock open: {interlock_open}")

    control_mode_local = module.control_mode_local
    print(f"Control mode local: {control_mode_local}")

    control_mode_remote = module.control_mode_remote
    print(f"Control mode remote: {control_mode_remote}")

    local_bus_termination_status = module.local_bus_termination_status
    print(f"Local bus termination status: {local_bus_termination_status}")

    local_bus_termination_status_on = module.local_bus_termination_status_on
    print(f"Local bus termination status on: {local_bus_termination_status_on}")

    local_bus_termination_status_off = module.local_bus_termination_status_off
    print(f"Local bus termination status off: {local_bus_termination_status_off}")

    board_alarm_status = module.board_alarm_status
    print(f"Board alarm status: {board_alarm_status}")

    channels = module.channels
    print(f"Channels: {channels}")


@serial_skip_decorator
def test_caen_channel_serial():
    caen = Caen(
        port=serial_port,
        baudrate=serial_baud,
        connect=True,
        timeout=timeout,
        logging_level=logging.DEBUG,
    )
    print(
        f"Serial port status: connected: {caen.connected}, port: {caen.port}, baudrate: {caen.baudrate}, timeout: {caen.timeout}"
    )
    module = caen.module(0)

    for channel in module.channels:
        print("")
        print(f"Channel: {channel.channel}")

        vmon = channel.vmon
        print(f"vmon: {vmon}")

        vset = channel.vset
        print(f"vset: {vset}")

        vmin = channel.vmin
        print(f"vmin: {vmin}")

        vmax = channel.vmax
        print(f"vmax: {vmax}")

        vdec = channel.vdec
        print(f"vdec: {vdec}")

        iset = channel.iset
        print(f"iset: {iset}")

        imin = channel.imin
        print(f"imin: {imin}")

        imax = channel.imax
        print(f"imax: {imax}")

        isdec = channel.isdec
        print(f"isdec: {isdec}")

        imon = channel.imon
        print(f"imon: {imon}")

        imrange = channel.imrange
        print(f"imrange: {imrange}")

        imdec = channel.imdec
        print(f"imdec: {imdec}")

        maxv = channel.maxv
        print(f"maxv: {maxv}")

        mvmin = channel.mvmin
        print(f"mvmin: {mvmin}")

        mvmax = channel.mvmax
        print(f"mvmax: {mvmax}")

        mvdec = channel.mvdec
        print(f"mvdec: {mvdec}")

        rup = channel.rup
        print(f"rup: {rup}")

        rupmin = channel.rupmin
        print(f"rupmin: {rupmin}")

        rupmax = channel.rupmax
        print(f"rupmax: {rupmax}")

        rupdec = channel.rupdec
        print(f"rupdec: {rupdec}")

        rdw = channel.rdw
        print(f"rdw: {rdw}")

        rdwmin = channel.rdwmin
        print(f"rdwmin: {rdwmin}")

        rdwmax = channel.rdwmax
        print(f"rdwmax: {rdwmax}")

        rdwdec = channel.rdwdec
        print(f"rdwdec: {rdwdec}")

        trip = channel.trip
        print(f"trip: {trip}")

        tripmin = channel.tripmin
        print(f"tripmin: {tripmin}")

        tripmax = channel.tripmax
        print(f"tripmax: {tripmax}")

        tripdec = channel.tripdec
        print(f"tripdec: {tripdec}")

        pdwn = channel.pdwn
        print(f"pdwn: {pdwn}")

        pol = channel.pol
        print(f"pol: {pol}")

        stat = channel.stat
        print(f"stat: {stat}")



from __future__ import annotations

# Dictionary mapping monitor channel commands to their descriptions
_mon_channel_commands = {
    "VSET": "Read out VSET value (XXXX.X V)",
    "VMIN": "Read out VSET minimum value (0 V)",
    "VMAX": "Read out VSET maximum value (8000.0 V)",
    "VDEC": "Read out VSET number of decimal digits",
    "VMON": "Read out VMON value (XXXX.X V)",
    "ISET": "Read out ISET value (XXXX.XX µA)",
    "IMIN": "Read out ISET minimum value (0 µA)",
    "IMAX": "Read out ISET maximum value (3000.00 µA)",
    "ISDEC": "Read out ISET number of decimal digits",
    "IMON": "Read out IMON value (XXXX.XX µA)",
    "IMRANGE": "Read out IMON RANGE value (HIGH / LOW)",
    "IMDEC": "Read out IMON number of decimal digits (2 HR, 3 LR)",
    "MAXV": "Read out MAXVSET value (XXXX V)",
    "MVMIN": "Read out MAXVSET minimum value (0 V)",
    "MVMAX": "Read out MAXVSET maximum value (8100 V)",
    "MVDEC": "Read out MAXVSET number of decimal digits",
    "RUP": "Read out RAMP UP value (XXX V/S)",
    "RUPMIN": "Read out RAMP UP minimum value (1 V/S)",
    "RUPMAX": "Read out RAMP UP maximum value (500 V/S)",
    "RUPDEC": "Read out RAMP UP number of decimal digits",
    "RDW": "Read out RAMP DOWN value (XXX V/S)",
    "RDWMIN": "Read out RAMP DOWN minimum value (1 V/S)",
    "RDWMAX": "Read out RAMP DOWN maximum value (500 V/S)",
    "RDWDEC": "Read out RAMP DOWN number of decimal digits",
    "TRIP": "Read out TRIP time value (XXXX.X S)",
    "TRIPMIN": "Read out TRIP time minimum value (0 S)",
    "TRIPMAX": "Read out TRIP time maximum value (1000.0 S)",
    "TRIPDEC": "Read out TRIP time number of decimal digits",
    "PDWN": "Read out POWER DOWN value (RAMP / KILL)",
    "POL": "Read out POLARITY value ('+' / '-')",
    "STAT": "Read out Channel status value (XXXXX)",
}

# Dictionary mapping set channel commands to their descriptions
_set_channel_commands = {
    "VSET": "VAL:XXXX.X Set VSET value",
    "ISET": "VAL:XXXX.XX Set ISET value",
    "MAXV": "VAL:XXXX Set MAXVSET value",
    "RUP": "VAL:XXX Set RAMP UP value",
    "RDW": "VAL:XXX Set RAMP DOWN value",
    "TRIP": "VAL:XXXX.X Set TRIP time value",
    "PDWN": "VAL:RAMP/KILL Set POWER DOWN mode value",
    "IMRANGE": "VAL:HIGH/LOW Set IMON RANGE value",
    "ON": "Set Ch ON",
    "OFF": "Set Ch OFF",
}


def validate_channel_number(channel: int) -> None:
    if not 0 <= channel <= 7:
        raise ValueError(
            f"Invalid channel number '{channel}'. Must be in the range 0..7."
        )


def validate_board_number(board: int) -> None:
    if not 0 <= board <= 31:
        raise ValueError(f"Invalid board number '{board}'. Must be in the range 0..31.")


def validate_channel_command(command: str, valid_commands: dict) -> None:
    if command.upper() not in valid_commands:
        valid_commands = ", ".join(valid_commands.keys())
        raise ValueError(
            f"Invalid command '{command}'. Valid command are: {valid_commands}"
        )


def _get_mon_channel_command(bd: int, channel: int, command: str) -> bytes:
    """
    Generate a command string to monitor a specific channel command.

    Args:
        bd (int): The board number.
        channel (int): The channel number.
        command (str): The command to monitor.


    Returns:
        bytes: The command string encoded as bytes.

    Raises:
        ValueError: If the provided command is not valid.
    """

    validate_board_number(bd)
    validate_channel_number(channel)

    command = command.upper()

    validate_channel_command(command, _mon_channel_commands)

    return f"$BD:{bd:02d},CMD:MON,CH:{channel:01d},PAR:{command}\r\n".encode("utf-8")


def _get_set_channel_command(
    bd: int, channel: int, command: str, value: str | int | float | None
) -> bytes:
    """
    Generate a command string to set a specific channel command to a given value.

    Args:
        bd (int): The board number.
        channel (int): The channel number.
        command (str): The command to set.
        value (str | int | float | None): The value to set the command to.

    Returns:
        bytes: The command string encoded as bytes.

    Raises:
        ValueError: If the provided command or value is not valid.
    """

    validate_board_number(bd)
    validate_channel_number(channel)

    command = command.upper()

    validate_channel_command(command, _set_channel_commands)

    if command in ["ON", "OFF"]:
        if value is not None:
            raise ValueError(f"command '{command}' does not accept a value")
        return f"$BD:{bd:02d},CMD:SET,CH:{channel:01d},PAR:{command}\r\n".encode(
            "utf-8"
        )
    elif command == "IMRANGE":
        if value not in ["HIGH", "LOW"]:
            raise ValueError(f"command '{command}' only accepts 'HIGH' or 'LOW'")
    elif command == "PDWN":
        if value not in ["RAMP", "KILL"]:
            raise ValueError(f"command '{command}' only accepts 'RAMP' or 'KILL'")

    return (
        f"$BD:{bd:02d},CMD:SET,CH:{channel:01d},PAR:{command},VAL:{value}\r\n".encode(
            "utf-8"
        )
    )



from __future__ import annotations
import re

import logging
import serial


def _write_command(
    ser: serial.Serial,
    logger: logging.Logger,
    bd: int,
    command: bytes,
    response: bool = True,
) -> str | None:
    logger.debug(f"Sending command: {command}")
    if not ser.is_open:
        logger.error("Serial port is not open")
        raise serial.SerialException("Serial port is not open")

    ser.write(command)
    if not response:
        logger.warning(
            "Calling _write_command without expecting a response. Manual readout of the response is required."
        )
        return None

    response = ser.readline()
    logger.debug(f"Received response: {response}")
    bd_from_response, response_value = _parse_response(response)
    if bd_from_response != bd:
        raise ValueError(
            f"Invalid response: {response_value}. Expected board number {bd}, got {bd_from_response}"
        )

    return response_value


def _parse_response(response: bytes) -> (int, str):
    """Parse the response from a device.

    Args:
        response (bytes): The response received from the device. (e.g. b"#BD:01,CMD:OK,VAL:42\r\n").

    Returns:
        (int, str): The board number and the value of the response.

    Raises:
        ValueError: If the response is invalid, cannot be decoded, or does not match the expected pattern.
    """

    if response == b"":
        raise ValueError(
            "Empty response. There was no response from the device. Check that the device is connected and correctly "
            "configured (baudrate)."
        )

    try:
        response: str = response.decode("utf-8").strip()
    except UnicodeDecodeError:
        raise ValueError(f"Invalid response: {response}")

    regex = re.compile(r"^#BD:(\d{2}),CMD:OK(?:,VAL:(.+))?$")
    match = regex.match(response)
    if match is None:
        raise ValueError(f"Invalid response: '{response}'. Could not match regex")
    bd = int(match.group(1))
    value: str | None = match.group(2) if match.group(2) else None

    return bd, value



import pytest

from hvps.commands.caen import _parse_response
from hvps.commands.caen.channel import (
    _get_set_channel_command,
    _get_mon_channel_command,
)

from hvps.commands.caen.module import (
    _get_set_module_command,
    _get_mon_module_command,
)


def test_caen_module_get_commands():
    with pytest.raises(ValueError):
        # invalid parameter name
        _get_mon_module_command(0, "TEST")

    with pytest.raises(ValueError):
        # invalid board number
        _get_mon_module_command(-1, "BDNAME")

    command = _get_mon_module_command(0, "BDNAME")
    assert command == b"$BD:00,CMD:MON,PAR:BDNAME\r\n"


def test_caen_module_set_commands():
    with pytest.raises(ValueError):
        # invalid parameter name
        _get_set_module_command(0, "TEST", 10)

    with pytest.raises(ValueError):
        # invalid board number
        _get_set_module_command(-1, "BDILKM", 10)

    command = _get_set_module_command(0, "BDILKM", 10)
    assert command == b"$BD:00,CMD:SET,PAR:BDILKM,VAL:10\r\n"


def test_caen_channel_get_commands():
    with pytest.raises(ValueError):
        # invalid parameter name
        _get_mon_channel_command(0, 1, "TEST")

    with pytest.raises(ValueError):
        # invalid board number
        _get_mon_channel_command(-1, 1, "BDNAME")

    with pytest.raises(ValueError):
        # invalid channel number
        _get_mon_channel_command(0, -1, "BDNAME")

    command = _get_mon_channel_command(0, 1, "VSET")
    assert command == b"$BD:00,CMD:MON,CH:1,PAR:VSET\r\n"


def test_caen_channel_set_commands():
    with pytest.raises(ValueError):
        # invalid parameter name
        _get_set_channel_command(0, 1, "TEST", 10)

    with pytest.raises(ValueError):
        # invalid parameter value
        _get_set_channel_command(0, 1, "IMRANGE", "INVALID")

    with pytest.raises(ValueError):
        # invalid board number
        _get_mon_channel_command(-1, 1, "BDNAME")

    with pytest.raises(ValueError):
        # invalid channel number
        _get_mon_channel_command(0, -1, "BDNAME")

    command = _get_set_channel_command(0, 1, "ON", None)
    assert command == b"$BD:00,CMD:SET,CH:1,PAR:ON\r\n"

    command = _get_set_channel_command(0, 1, "IMRANGE", "LOW")
    assert command == b"$BD:00,CMD:SET,CH:1,PAR:IMRANGE,VAL:LOW\r\n"

    command = _get_set_channel_command(0, 1, "PDWN", "KILL")
    assert command == b"$BD:00,CMD:SET,CH:1,PAR:PDWN,VAL:KILL\r\n"


def test_caen_parse_response():
    response = b"#BD:01,CMD:OK,VAL:42\r\n"
    bd, value = _parse_response(response)
    assert bd == 1
    assert value == "42"

    response = b"#BD:99,CMD:OK,VAL:Hello World\r\n"
    bd, value = _parse_response(response)
    assert bd == 99
    assert value == "Hello World"

    with pytest.raises(ValueError):
        # Invalid response format
        response = b"Invalid response\r\n"
        _parse_response(response)

    response = b"#BD:09,CMD:OK\r\n"
    bd, value = _parse_response(response)
    assert bd == 9
    assert value is None

    with pytest.raises(ValueError):
        # number needs to be two digits
        response = b"#BD:9,CMD:OK\r\n"
        _parse_response(response)

    with pytest.raises(ValueError):
        # add additional ','
        response = b"#BD:9,CMD:OK,\r\n"
        _parse_response(response)



from __future__ import annotations

# Dictionary mapping monitor module commands to their descriptions
_mon_module_commands = {
    "BDNAME": "Read out module name (N1471)",
    "BDNCH": "Read out number of Channels present (4)",
    "BDFREL": "Read out Firmware Release (XX.X)",
    "BDSNUM": "Read out value serial number (XXXXX)",
    "BDILK": "Read out INTERLOCK status (YES/NO)",
    "BDILKM": "Read out INTERLOCK mode (OPEN/CLOSED)",
    "BDCTR": "Read out Control Mode (LOCAL/REMOTE)",
    "BDTERM": "Read out LOCAL BUS Termination status (ON/OFF)",
    "BDALARM": "Read out Board Alarm status value (XXXXX)",
}

# Dictionary mapping set module commands to their descriptions
_set_module_commands = {
    "BDILKM": "VAL:OPEN/CLOSED Set Interlock Mode",
    "BDCLR": "Clear alarm signal",
}


def _get_mon_module_command(bd: int, command: str) -> bytes:
    """
    Generate a command string to monitor a specific module command.

    Args:
        bd (int): The board number. Must be in the range 0..31.
        command (str): The command to monitor.

    Returns:
        bytes: The command string encoded as bytes.

    Raises:
        ValueError: If the provided command is not valid.
    """
    if not 0 <= bd <= 31:
        raise ValueError(f"Invalid board number '{bd}'. Must be in the range 0..31.")

    command = command.upper()
    if command not in _mon_module_commands:
        valid_commands = ", ".join(_mon_module_commands.keys())
        raise ValueError(
            f"Invalid command '{command}'. Valid commands are: {valid_commands}"
        )
    return f"$BD:{bd:02d},CMD:MON,PAR:{command}\r\n".encode("utf-8")


def _get_set_module_command(
    bd: int, command: str, value: str | int | float | None
) -> bytes:
    """
    Generate a command string to set a specific module command to a given value.

    Args:
        bd (int): The board number.
        command (str): The command to set.
        value (str | int | float): The value to set the command to.

    Returns:
        bytes: The command string encoded as bytes.

    Raises:
        ValueError: If the provided command or value is not valid.
    """
    if not 0 <= bd <= 31:
        raise ValueError(f"Invalid board number '{bd}'. Must be in the range 0..31.")

    command = command.upper()
    if command not in _set_module_commands:
        valid_commands = ", ".join(_set_module_commands.keys())
        raise ValueError(
            f"Invalid command '{command}'. Valid commands are: {valid_commands}"
        )
    return f"$BD:{bd:02d},CMD:SET,PAR:{command},VAL:{value}\r\n".encode("utf-8")



from ...commands.caen.channel import (
    _get_set_channel_command,
    _get_mon_channel_command,
)
from ...commands.caen import _write_command
from ...utils import string_number_to_bit_array
from ..channel import Channel as BaseChannel

from time import sleep


class Channel(BaseChannel):
    def __init__(self, *args, bd: int, **kwargs):
        super().__init__(*args, **kwargs)
        self._bd = bd

    @property
    def bd(self) -> int:
        """The bd value of the channel.

        Returns:
            int: The bd value.
        """
        return self._bd

    def wait_for_vset(self, timeout: float = 60.0):
        """Wait for the vset value to stabilize within a specified voltage difference.

        Args:
            timeout (float, optional): The maximum time to wait in seconds. Defaults to 60.0 seconds.

        Raises:
            TimeoutError: If the vset value does not stabilize within the specified timeout.

        """
        timedelta = 0.5  # 0.5 seconds
        t = 0.0
        while t < timeout:
            t += timedelta
            sleep(timedelta)
            if self.voltage_target_reached:
                return

        # Could not stabilize within the specified timeout
        raise TimeoutError(f"Could not stabilize vset within {timeout} seconds.")

    # Getters
    @property
    def vset(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "VSET"),
        )
        return float(response)

    @property
    def voltage_set(self) -> float:
        return self.vset

    @property
    def vmin(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "VMIN"),
        )
        return float(response)

    @property
    def vmax(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "VMAX"),
        )
        return float(response)

    @property
    def vdec(self) -> int:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "VDEC"),
        )
        return int(response)

    @property
    def vmon(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "VMON"),
        )
        return float(response)

    @property
    def iset(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "ISET"),
        )
        return float(response)

    @property
    def imin(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "IMIN"),
        )
        return float(response)

    @property
    def imax(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "IMAX"),
        )
        return float(response)

    @property
    def isdec(self) -> int:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "ISDEC"),
        )
        return int(response)

    @property
    def imon(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "IMON"),
        )
        return float(response)

    @property
    def imrange(self) -> bool:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "IMRANGE"),
        )
        if response not in ["HIGH", "LOW"]:
            raise ValueError(f"Unexpected response {response}")
        return response == "HIGH"

    @property
    def imdec(self) -> int:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "IMDEC"),
        )
        return int(response)

    @property
    def maxv(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "MAXV"),
        )
        return float(response)

    @property
    def mvmin(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "MVMIN"),
        )
        return float(response)

    @property
    def mvmax(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "MVMAX"),
        )
        return float(response)

    @property
    def mvdec(self) -> int:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "MVDEC"),
        )
        return int(response)

    @property
    def rup(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "RUP"),
        )
        return float(response)

    @property
    def rupmin(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "RUPMIN"),
        )
        return float(response)

    @property
    def rupmax(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "RUPMAX"),
        )
        return float(response)

    @property
    def rupdec(self) -> int:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "RUPDEC"),
        )
        return int(response)

    @property
    def rdw(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "RDW"),
        )
        return float(response)

    @property
    def rdwmin(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "RDWMIN"),
        )
        return float(response)

    @property
    def rdwmax(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "RDWMAX"),
        )
        return float(response)

    @property
    def rdwdec(self) -> int:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "RDWDEC"),
        )
        return int(response)

    @property
    def trip(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "TRIP"),
        )
        return float(response)

    @property
    def tripmin(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "TRIPMIN"),
        )
        return float(response)

    @property
    def tripmax(self) -> float:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "TRIPMAX"),
        )
        return float(response)

    @property
    def tripdec(self) -> int:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "TRIPDEC"),
        )
        return int(response)

    @property
    def pdwn(self) -> str:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "PDWN"),
        )
        return str(response)

    @property
    def pol(self) -> str:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "POL"),
        )
        if response not in ["+", "-"]:
            raise ValueError(f"Invalid polarity: {response}")
        return str(response)

    def polarity_positive(self) -> bool:
        return self.pol == "+"

    def polarity_negative(self) -> bool:
        return self.pol == "-"

    @property
    def stat(self) -> dict:
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_mon_channel_command(self._bd, self._channel, "STAT"),
        )
        bit_array = string_number_to_bit_array(response)

        return {
            "ON": bit_array[0],  # True: ON, False: OFF
            "RUP": bit_array[1],  # True: Channel Ramp UP
            "RDW": bit_array[2],  # True: Channel Ramp DOWN
            "OVC": bit_array[3],  # True: IMON >= ISET
            "OVV": bit_array[4],  # True: VMON > VSET + 2.5 V
            "UNV": bit_array[5],  # True: VMON < VSET – 2.5 V
            "MAXV": bit_array[6],  # True: VOUT in MAXV protection
            "TRIP": bit_array[7],  # True: Ch OFF via TRIP (Imon >= Iset during TRIP)
            "OVP": bit_array[8],  # True: Output Power > Max
            "OVT": bit_array[9],  # True: TEMP > 105°C
            "DIS": bit_array[
                10
            ],  # True: Ch disabled (REMOTE Mode and Switch on OFF position)
            "KILL": bit_array[11],  # True: Ch in KILL via front panel
            "ILK": bit_array[12],  # True: Ch in INTERLOCK via front panel
            "NOCAL": bit_array[13],  # True: Calibration Error
            # "NC": bit_array[14]  # True: Not Connected
        }

    @property
    def voltage_target_reached(self) -> bool:
        stat = self.stat
        return not stat["OVV"] and not stat["UNV"]

    @property
    def on(self) -> bool:
        return self.stat["ON"]

    @property
    def off(self) -> bool:
        return not self.on

    @property
    def kill(self) -> bool:
        return self.stat["KILL"]

    # Setters
    @vset.setter
    def vset(self, value: float) -> None:
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "VSET", value),
        )
        if self.vset != value:
            raise ValueError(f"Could not set VSET to {value}")

    @iset.setter
    def iset(self, value: float) -> None:
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "ISET", value),
        )
        if self.iset != value:
            raise ValueError(f"Could not set ISET to {value}")

    @maxv.setter
    def maxv(self, value: float) -> None:
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "MAXV", value),
        )
        if self.maxv != value:
            raise ValueError(f"Could not set MAXV to {value}")

    @rup.setter
    def rup(self, value: float) -> None:
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "RUP", value),
        )
        if self.rup != value:
            raise ValueError(f"Could not set RUP to {value}")

    @rdw.setter
    def rdw(self, value: float) -> None:
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "RDW", value),
        )
        if self.rdw != value:
            raise ValueError(f"Could not set RDW to {value}")

    @trip.setter
    def trip(self, value: float) -> None:
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "TRIP", value),
        )
        if self.trip != value:
            raise ValueError(f"Could not set TRIP to {value}")

    @pdwn.setter
    def pdwn(self, value: float) -> None:
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "PDWN", value),
        )
        if self.pdwn != value:
            raise ValueError(f"Could not set PDWN to {value}")

    @imrange.setter
    def imrange(self, value: float) -> None:
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "IMRANGE", value),
        )
        if self.imrange != value:
            raise ValueError(f"Could not set IMRANGE to {value}")

    def turn_on(self) -> None:
        """Turn on the channel."""
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "ON", None),
        )

    def turn_off(self) -> None:
        """Turn off the channel."""
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self._bd,
            command=_get_set_channel_command(self._bd, self._channel, "OFF", None),
        )



from hvps import Iseg
import pytest


def test_iseg_init(caplog):
    caplog.set_level("DEBUG")

    Iseg(connect=False)

    assert caplog.text == ""

    iseg = Iseg(connect=False, logging_level="DEBUG")

    assert iseg.baudrate == 115200
    assert "Using baud rate 115200" in caplog.text
    assert "Using port " in caplog.text
    assert "Using timeout " in caplog.text


def test_iseg_module(caplog):
    iseg = Iseg(connect=False)

    # for ISEG only one module exists
    iseg.module()
    iseg.module(0)

    # check exception takes place
    with pytest.raises(Exception):
        iseg.module(-1)

    with pytest.raises(Exception):
        iseg.module(1)


def test_iseg_channel(caplog):
    caplog.set_level("DEBUG")

    iseg = Iseg(connect=False, logging_level="DEBUG")

    module = iseg.module()

    channel = module.channel(0)

    assert "Creating channel 0" in caplog.text

    print(f"channel: {channel.channel}")



import os
import subprocess
import sys


def run_main_with_arguments(arguments: list) -> tuple:
    main_file_path = os.path.join(
        os.path.dirname(__file__), "..", "src/hvps", "__main__.py"
    )
    main_file_path = os.path.abspath(main_file_path)

    print(f"__main__.py file path: {main_file_path}")

    process = subprocess.Popen(
        [sys.executable, main_file_path] + arguments,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()
    exit_code = process.returncode

    return stdout.decode(), stderr.decode(), exit_code


def test_cli_channel_mon():
    for parameter in [
        "VSET",
        "VMIN",
        "VMAX",
        "VDEC",
        "VMON",
        "ISET",
        "IMIN",
        "IMAX",
        "ISDEC",
        "IMON",
        "IMRANGE",
        "IMDEC",
        "MAXV",
        "MVMIN",
        "MVMAX",
        "MVDEC",
        "RUP",
        "RUPMIN",
        "RUPMAX",
        "RUPDEC",
        "RDW",
        "RDWMIN",
        "RDWMAX",
        "RDWDEC",
        "TRIP",
        "TRIPMIN",
        "TRIPMAX",
        "TRIPDEC",
        "PDWN",
        "POL",
        "STAT",
        "INVALID_PARAMETER",
    ]:
        arguments = [
            "--port",
            "/dev/ttyUSB0",
            "--baud",
            "9600",
            "--module",
            "1",
            "--channel",
            "2",
            "--test",
            "caen",
            f"{parameter}",
        ]

        stdout, stderr, exit_code = run_main_with_arguments(arguments)

        print(f"arguments: {arguments}")
        print(f"stdout: {stdout}")
        print(f"stderr: {stderr}")
        print(f"exit_code: {exit_code}")

        exit_code_expected = 0 if parameter not in ["INVALID_PARAMETER"] else 1
        assert (
            exit_code == exit_code_expected
        ), f"exit_code: {exit_code} for arguments: {arguments}"


def test_cli_channel_set():
    for parameter in [
        "VSET",
        "ISET",
        "MAXV",
        "RUP",
        "RDW",
        "TRIP",
        "PDWN",
        "IMRANGE",
        "ON",
        "OFF",
    ]:
        arguments = [
            "--port",
            "/dev/ttyUSB0",
            "--baud",
            "9600",
            "--module",
            "1",
            "--channel",
            "2",
            "--test",
            "caen",
            f"{parameter}",
            "100",
        ]

        stdout, stderr, exit_code = run_main_with_arguments(arguments)

        print(f"arguments: {arguments}")
        print(f"stdout: {stdout}")
        print(f"stderr: {stderr}")
        print(f"exit_code: {exit_code}")

        # ON / OFF are special cases, they cannot be set to a value, so they should return exit code 1
        exit_code_expected = 0 if parameter not in ["ON", "OFF"] else 1
        assert (
            exit_code == exit_code_expected
        ), f"exit_code: {exit_code} for arguments: {arguments}"


def test_cli_module_mon():
    for parameter in [
        "BDNAME",
        "BDNCH",
        "BDFREL",
        "BDSNUM",
        "BDILK",
        "BDILKM",
        "BDCTR",
        "BDTERM",
        "BDALARM",
        "INVALID_PARAMETER",
    ]:
        arguments = [
            "--port",
            "/dev/ttyUSB0",
            "--baud",
            "9600",
            "--module",
            "1",
            "--test",
            "caen",
            f"{parameter}",
        ]

        stdout, stderr, exit_code = run_main_with_arguments(arguments)

        print(f"arguments: {arguments}")
        print(f"stdout: {stdout}")
        print(f"stderr: {stderr}")
        print(f"exit_code: {exit_code}")

        exit_code_expected = 0 if parameter not in ["INVALID_PARAMETER"] else 1
        assert (
            exit_code == exit_code_expected
        ), f"exit_code: {exit_code} for arguments: {arguments}"


def test_cli_module_set():
    for parameter in [
        "BDILKM",
        "BDCLR",
    ]:
        arguments = [
            "--port",
            "/dev/ttyUSB0",
            "--baud",
            "9600",
            "--module",
            "2",
            "--test",
            "caen",
            f"{parameter}",
            "200",
        ]

        stdout, stderr, exit_code = run_main_with_arguments(arguments)

        print(f"arguments: {arguments}")
        print(f"stdout: {stdout}")
        print(f"stderr: {stderr}")
        print(f"exit_code: {exit_code}")

        # ON / OFF are special cases, they cannot be set to a value, so they should return exit code 1
        exit_code_expected = 0 if parameter not in ["ON", "OFF"] else 1
        assert (
            exit_code == exit_code_expected
        ), f"exit_code: {exit_code} for arguments: {arguments}"



from serial.tools import list_ports
from typing import List


def string_number_to_bit_array(string) -> List[bool]:
    """
    Converts a string representing a 32-bit integer into a list of bits.

    Args:
        string (str): The string to convert. It must be a string representing an integer (e.g. "01024").

    Returns:
        list: The list of 32 bits, from least significant to most significant.
    """

    try:
        string_as_int = int(string)
    except ValueError:
        raise ValueError(f"Invalid string '{string}'. Must be an integer.")

    return list(reversed([bool(int(bit)) for bit in f"{string_as_int:032b}"]))


def get_serial_ports() -> List[str]:
    """
    Get a list of available serial ports.

    Returns:
        list: A list of available serial ports.
    """
    return [port.device for port in list_ports.comports()]



from functools import cached_property
from typing import List

from ...commands.caen.module import _get_mon_module_command, _get_set_module_command
from ...commands.caen import _write_command
from ...utils.utils import string_number_to_bit_array
from .channel import Channel
from ..module import Module as BaseModule


class Module(BaseModule):
    @property
    def bd(self) -> int:
        """The bd value of the module.

        Returns:
            int: The bd value.
        """
        return self.module

    def channel(self, channel: int) -> Channel:
        return super().channel(channel)

    @cached_property
    def number_of_channels(self) -> int:
        """The number of channels in the module.

        Returns:
            int: The number of channels.
        """
        self._logger.debug("Getting number of channels")
        if not self._serial.is_open:
            # TODO: we should not cache the property if the serial port is not open
            self._logger.warning(
                "Serial port is not open. Returning 1 as number of channels."
            )
            return 1

        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_mon_module_command(self.bd, "BDNCH"),
        )
        return int(response)

    @property
    def channels(self) -> List[Channel]:
        """The channels in the module.

        Returns:
            List[Channel]: A list of Channel objects.
        """
        if len(self._channels) == 0:
            self._logger.debug("Initializing channels")
            for channel in range(self.number_of_channels):
                self._logger.debug(f"Creating channel {channel}")
                self._channels.append(
                    Channel(
                        ser=self._serial,
                        logger=self._logger,
                        channel=channel,
                        bd=self.bd,
                    )
                )
        return self._channels

    @cached_property
    def name(self) -> str:
        """The name of the module.

        Returns:
            str: The name of the module.
        """
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_mon_module_command(self.bd, "BDNAME"),
        )
        return str(response)

    @property
    def firmware_release(self) -> str:
        """
        Read out Firmware Release (XX.X)

        Returns:
            str: The firmware release.
        """
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_mon_module_command(self.bd, "BDFREL"),
        )
        return response

    @property
    def serial_number(self) -> str:
        """
        Read out value serial number (XXXXX)

        Returns:
            str: The serial number.
        """
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_mon_module_command(self.bd, "BDSNUM"),
        )
        # TODO: check if can be casted to int
        return response

    @property
    def interlock_status(self) -> bool:
        """
        Read out INTERLOCK status (YES/NO)

        Returns:
            bool: Yes: True, No: False
        """
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_mon_module_command(self.bd, "BDILK"),
        )
        if response == "YES":
            return True
        elif response == "NO":
            return False
        else:
            raise ValueError(f"Invalid response for 'interlock_status': {response}")

    @property
    def interlock_mode(self) -> str:
        """
        Read out INTERLOCK mode (OPEN/CLOSED)

        Returns:
            str: The interlock mode.
        """
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_mon_module_command(self.bd, "BDILKM"),
        )
        if response not in ["OPEN", "CLOSED"]:
            raise ValueError(f"Invalid response for 'interlock_mode': {response}")
        return response

    @property
    def interlock_open(self) -> bool:
        """
        Returns:
            bool: True if interlock_mode == OPEN
        """
        return self.interlock_mode == "OPEN"

    @property
    def control_mode(self) -> str:
        """
        Read out Control Mode (LOCAL / REMOTE)

        Returns:
            str: The control mode.
        """
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_mon_module_command(self.bd, "BDCTR"),
        )
        if response not in ["LOCAL", "REMOTE"]:
            raise ValueError(f"Invalid response for 'control_mode': {response}")
        return response

    @property
    def control_mode_local(self) -> bool:
        """
        Returns:
            bool: True if control_mode == LOCAL
        """
        return self.control_mode == "LOCAL"

    @property
    def control_mode_remote(self) -> bool:
        """
        Returns:
            bool: True if control_mode == REMOTE
        """
        return self.control_mode == "REMOTE"

    @control_mode.setter
    def control_mode(self, value: str):
        """
        Set Control Mode (LOCAL / REMOTE)

        Args:
            value (str): The control mode.
        """
        value = value.upper()
        if value not in ["LOCAL", "REMOTE"]:
            raise ValueError(f"Invalid value for 'control_mode': {value}")

        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_set_module_command(self.bd, "BDCTR", value),
        )

    def set_control_mode_local(self):
        """
        Set Control Mode to LOCAL
        """
        self.control_mode = "LOCAL"

    def set_control_mode_remote(self):
        """
        Set Control Mode to REMOTE
        """
        self.control_mode = "REMOTE"

    @property
    def local_bus_termination_status(self) -> str:
        """
        Read out LOCAL BUS Termination status (ON/OFF)

        Returns:
            str: The local bus termination status.
        """
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_mon_module_command(self.bd, "BDTERM"),
        )
        if response not in ["ON", "OFF"]:
            raise ValueError(
                f"Invalid response for 'local_bus_termination_status': {response}"
            )
        return response

    @property
    def local_bus_termination_status_on(self) -> bool:
        """
        Returns:
            bool: True if local_bus_termination_status == ON
        """
        return self.local_bus_termination_status == "ON"

    @property
    def local_bus_termination_status_off(self) -> bool:
        """
        Returns:
            bool: True if local_bus_termination_status == OFF
        """
        return self.local_bus_termination_status == "OFF"

    @property
    def board_alarm_status(self) -> dict:
        """
        Read out Board Alarm status value (XXXXX)

        Returns:
            str: The board alarm status value.
        """
        response = _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_mon_module_command(self.bd, "BDALARM"),
        )

        bit_array = string_number_to_bit_array(response)

        return {
            "CH0": bit_array[0],  # True: Ch0 in Alarm status
            "CH1": bit_array[1],  # True: Ch1 in Alarm status
            "CH2": bit_array[2],  # True: Ch2 in Alarm status
            "CH3": bit_array[3],  # True: Ch3 in Alarm status
            "PWFAIL": bit_array[4],  # True: Board in POWER FAIL
            "OVP": bit_array[5],  # True: Board in OVER POWER
            "HVCKFAIL": bit_array[6],  # True: Internal HV Clock FAIL (≠ 200±10kHz)
        }

    @interlock_mode.setter
    def interlock_mode(self, mode: str) -> None:
        """
        VAL:OPEN/CLOSED Set Interlock Mode

        Args:
            mode (str): The interlock mode to set.
        """
        mode = mode.upper()
        if mode not in ["OPEN", "CLOSED"]:
            raise ValueError(f"Invalid value for 'interlock_mode': {mode}")

        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_set_module_command(self.bd, "BDILKM", mode),
        )

    def close_interlock(self) -> None:
        """
        Close Interlock
        """
        self.interlock_mode = "CLOSED"

    def open_interlock(self) -> None:
        """
        Open Interlock
        """
        self.interlock_mode = "OPEN"

    def clear_alarm_signal(self) -> None:
        """
        Clear alarm signal
        """
        _write_command(
            ser=self._serial,
            logger=self._logger,
            bd=self.bd,
            command=_get_set_module_command(self.bd, "BDCLR", None),
        )



from __future__ import annotations

from ..hvps import Hvps
from .module import Module


class Iseg(Hvps):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._modules = {0: Module(self._serial, self._logger, 0)}



import serial
from abc import ABC, abstractmethod
import logging


class Channel(ABC):
    def __init__(self, ser: serial.Serial, logger: logging.Logger, channel: int):
        """Initialize the Channel object.

        Args:
            ser (serial.Serial): The serial object used for communication.
            logger (logging.Logger): The logger object used for logging.
            channel (int): The channel number.

        """
        self._serial = ser
        self._logger = logger
        self._channel = channel

    @property
    def channel(self) -> int:
        """The channel number.

        Returns:
            int: The channel number.
        """
        return self._channel

    @property
    @abstractmethod
    def voltage_set(self) -> float:
        """The set voltage of the channel.

        This property should be implemented by subclasses to provide the set voltage of the channel.

        Returns:
            float: The set voltage.

        Raises:
            NotImplementedError: If the subclass does not implement this property.

        """
        pass



import re

input_file = "module.py"
output_file = "output.py"

pattern1 = r"-> ([^:]*):"
pattern2 = r"_write_command\(self._serial, command"
group1 = ""

with open(input_file, "r") as input_f, open(output_file, "w") as output_f:
    for line in input_f:
        match = re.search(pattern1, line)
        if match:
            group1 = match.group(1)  # Extract the wildcard value

        modified_line = re.sub(
            pattern2, f"_write_command(self._serial, command, {group1}", line
        )
        output_f.write(modified_line)



from hvps import Caen
import pytest


def test_caen_init(caplog):
    caplog.set_level("DEBUG")

    Caen(connect=False)

    assert caplog.text == ""

    caen = Caen(connect=False, logging_level="DEBUG")

    assert caen.baudrate == 115200
    assert "Using baud rate 115200" in caplog.text
    assert "Using port " in caplog.text
    assert "Using timeout " in caplog.text


def test_caen_module(caplog):
    caplog.set_level("DEBUG")

    caen = Caen(connect=False, logging_level="DEBUG")

    # for CAEN, modules are dynamically created
    [caen.module(i) for i in range(0, 32)]

    assert "Getting module 0" in caplog.text

    # check exception takes place
    with pytest.raises(Exception):
        caen.module(-1)

    with pytest.raises(Exception):
        caen.module(35)


def test_caen_channel(caplog):
    caplog.set_level("DEBUG")

    caen = Caen(connect=False, logging_level="DEBUG")

    module = caen.module()

    channel = module.channel(0)

    assert "Creating channel 0" in caplog.text

    print(f"channel: {channel.channel}")
